# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""argument_action.py"""
from __future__ import annotations

from enum import Enum


class ArgumentAction(Enum):
    """Defines the action to be taken when the argument is encountered."""

    ACTION = "action"
    STORE = "store"
    STORE_TRUE = "store_true"
    STORE_FALSE = "store_false"
    STORE_BOOL_OPTIONAL = "store_bool_optional"
    APPEND = "append"
    EXTEND = "extend"
    COUNT = "count"
    HELP = "help"

    @classmethod
    def choices(cls) -> list[ArgumentAction]:
        """Return a list of all argument actions."""
        return list(cls)

    def __str__(self) -> str:
        """Return the string representation of the argument action."""
        return self.value



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
import types
from datetime import datetime
from enum import EnumMeta
from typing import Any, Literal, Union, get_args, get_origin

from dateutil import parser as date_parser

from falyx.action.base_action import BaseAction
from falyx.logger import logger
from falyx.parser.signature import infer_args_from_func


def coerce_bool(value: str) -> bool:
    if isinstance(value, bool):
        return value
    value = value.strip().lower()
    if value in {"true", "t", "1", "yes", "on"}:
        return True
    elif value in {"false", "f", "0", "no", "off"}:
        return False
    return bool(value)


def coerce_enum(value: Any, enum_type: EnumMeta) -> Any:
    if isinstance(value, enum_type):
        return value

    if isinstance(value, str):
        try:
            return enum_type[value]
        except KeyError:
            pass

    base_type = type(next(iter(enum_type)).value)
    try:
        coerced_value = base_type(value)
        return enum_type(coerced_value)
    except (ValueError, TypeError):
        values = [str(enum.value) for enum in enum_type]
        raise ValueError(f"'{value}' should be one of {{{', '.join(values)}}}") from None


def coerce_value(value: str, target_type: type) -> Any:
    origin = get_origin(target_type)
    args = get_args(target_type)

    if origin is Literal:
        if value not in args:
            raise ValueError(
                f"Value '{value}' is not a valid literal for type {target_type}"
            )
        return value

    if isinstance(target_type, types.UnionType) or get_origin(target_type) is Union:
        for arg in args:
            try:
                return coerce_value(value, arg)
            except Exception:
                continue
        raise ValueError(f"Value '{value}' could not be coerced to any of {args}")

    if isinstance(target_type, EnumMeta):
        return coerce_enum(value, target_type)

    if target_type is bool:
        return coerce_bool(value)

    if target_type is datetime:
        try:
            return date_parser.parse(value)
        except ValueError as e:
            raise ValueError(f"Value '{value}' could not be parsed as a datetime") from e

    return target_type(value)


def same_argument_definitions(
    actions: list[Any],
    arg_metadata: dict[str, str | dict[str, Any]] | None = None,
) -> list[dict[str, Any]] | None:

    arg_sets = []
    for action in actions:
        if isinstance(action, BaseAction):
            infer_target, _ = action.get_infer_target()
            arg_defs = infer_args_from_func(infer_target, arg_metadata)
        elif callable(action):
            arg_defs = infer_args_from_func(action, arg_metadata)
        else:
            logger.debug("Auto args unsupported for action: %s", action)
            return None
        arg_sets.append(arg_defs)

    first = arg_sets[0]
    if all(arg_set == first for arg_set in arg_sets[1:]):
        return first
    return None



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""command_argument_parser.py"""
from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from typing import Any, Iterable

from rich.console import Console
from rich.markup import escape

from falyx.action.base_action import BaseAction
from falyx.console import console
from falyx.exceptions import CommandArgumentError
from falyx.parser.argument import Argument
from falyx.parser.argument_action import ArgumentAction
from falyx.parser.parser_types import false_none, true_none
from falyx.parser.utils import coerce_value
from falyx.signals import HelpSignal


class CommandArgumentParser:
    """
    Custom argument parser for Falyx Commands.
    It is used to create a command-line interface for Falyx
    commands, allowing users to specify options and arguments
    when executing commands.
    It is not intended to be a full-featured replacement for
    argparse, but rather a lightweight alternative for specific use
    cases within the Falyx framework.

    Features:
    - Customizable argument parsing.
    - Type coercion for arguments.
    - Support for positional and keyword arguments.
    - Support for default values.
    - Support for boolean flags.
    - Support for optional boolean flags.
    - Exception handling for invalid arguments.
    - Render Help using Rich library.
    """

    def __init__(
        self,
        command_key: str = "",
        command_description: str = "",
        command_style: str = "bold",
        help_text: str = "",
        help_epilog: str = "",
        aliases: list[str] | None = None,
    ) -> None:
        """Initialize the CommandArgumentParser."""
        self.console: Console = console
        self.command_key: str = command_key
        self.command_description: str = command_description
        self.command_style: str = command_style
        self.help_text: str = help_text
        self.help_epilog: str = help_epilog
        self.aliases: list[str] = aliases or []
        self._arguments: list[Argument] = []
        self._positional: dict[str, Argument] = {}
        self._keyword: dict[str, Argument] = {}
        self._keyword_list: list[Argument] = []
        self._flag_map: dict[str, Argument] = {}
        self._dest_set: set[str] = set()
        self._add_help()

    def _add_help(self):
        """Add help argument to the parser."""
        self.add_argument(
            "-h",
            "--help",
            action=ArgumentAction.HELP,
            help="Show this help message.",
            dest="help",
        )

    def _is_positional(self, flags: tuple[str, ...]) -> bool:
        """Check if the flags are positional."""
        positional = False
        if any(not flag.startswith("-") for flag in flags):
            positional = True

        if positional and len(flags) > 1:
            raise CommandArgumentError("Positional arguments cannot have multiple flags")
        return positional

    def _get_dest_from_flags(self, flags: tuple[str, ...], dest: str | None) -> str:
        """Convert flags to a destination name."""
        if dest:
            if not dest.replace("_", "").isalnum():
                raise CommandArgumentError(
                    "dest must be a valid identifier (letters, digits, and underscores only)"
                )
            if dest[0].isdigit():
                raise CommandArgumentError("dest must not start with a digit")
            return dest
        dest = None
        for flag in flags:
            if flag.startswith("--"):
                dest = flag.lstrip("-").replace("-", "_").lower()
                break
            elif flag.startswith("-"):
                dest = flag.lstrip("-").replace("-", "_").lower()
            else:
                dest = flag.replace("-", "_").lower()
        assert dest is not None, "dest should not be None"
        if not dest.replace("_", "").isalnum():
            raise CommandArgumentError(
                "dest must be a valid identifier (letters, digits, and underscores only)"
            )
        if dest[0].isdigit():
            raise CommandArgumentError("dest must not start with a digit")
        return dest

    def _determine_required(
        self,
        required: bool,
        positional: bool,
        nargs: int | str | None,
        action: ArgumentAction,
    ) -> bool:
        """Determine if the argument is required."""
        if required:
            if action in (
                ArgumentAction.STORE_TRUE,
                ArgumentAction.STORE_FALSE,
                ArgumentAction.STORE_BOOL_OPTIONAL,
                ArgumentAction.HELP,
            ):
                raise CommandArgumentError(
                    f"Argument with action {action} cannot be required"
                )
            return True
        if positional:
            assert (
                nargs is None
                or isinstance(nargs, int)
                or isinstance(nargs, str)
                and nargs in ("+", "*", "?")
            ), f"Invalid nargs value: {nargs}"
            if isinstance(nargs, int):
                return nargs > 0
            elif isinstance(nargs, str):
                if nargs in ("+"):
                    return True
                elif nargs in ("*", "?"):
                    return False
            else:
                return True

        return required

    def _validate_nargs(
        self, nargs: int | str | None, action: ArgumentAction
    ) -> int | str | None:
        if action in (
            ArgumentAction.STORE_FALSE,
            ArgumentAction.STORE_TRUE,
            ArgumentAction.COUNT,
            ArgumentAction.HELP,
            ArgumentAction.STORE_BOOL_OPTIONAL,
        ):
            if nargs is not None:
                raise CommandArgumentError(
                    f"nargs cannot be specified for {action} actions"
                )
            return None
        if nargs is None:
            return None
        allowed_nargs = ("?", "*", "+")
        if isinstance(nargs, int):
            if nargs <= 0:
                raise CommandArgumentError("nargs must be a positive integer")
        elif isinstance(nargs, str):
            if nargs not in allowed_nargs:
                raise CommandArgumentError(f"Invalid nargs value: {nargs}")
        else:
            raise CommandArgumentError(f"nargs must be an int or one of {allowed_nargs}")
        return nargs

    def _normalize_choices(
        self, choices: Iterable | None, expected_type: Any, action: ArgumentAction
    ) -> list[Any]:
        if choices is not None:
            if action in (
                ArgumentAction.STORE_TRUE,
                ArgumentAction.STORE_FALSE,
                ArgumentAction.STORE_BOOL_OPTIONAL,
            ):
                raise CommandArgumentError(
                    f"choices cannot be specified for {action} actions"
                )
            if isinstance(choices, dict):
                raise CommandArgumentError("choices cannot be a dict")
            try:
                choices = list(choices)
            except TypeError:
                raise CommandArgumentError(
                    "choices must be iterable (like list, tuple, or set)"
                )
        else:
            choices = []
        for choice in choices:
            try:
                coerce_value(choice, expected_type)
            except Exception as error:
                raise CommandArgumentError(
                    f"Invalid choice {choice!r}: not coercible to {expected_type.__name__} error: {error}"
                ) from error
        return choices

    def _validate_default_type(
        self, default: Any, expected_type: type, dest: str
    ) -> None:
        """Validate the default value type."""
        if default is not None:
            try:
                coerce_value(default, expected_type)
            except Exception as error:
                raise CommandArgumentError(
                    f"Default value {default!r} for '{dest}' cannot be coerced to {expected_type.__name__} error: {error}"
                ) from error

    def _validate_default_list_type(
        self, default: list[Any], expected_type: type, dest: str
    ) -> None:
        if isinstance(default, list):
            for item in default:
                try:
                    coerce_value(item, expected_type)
                except Exception as error:
                    raise CommandArgumentError(
                        f"Default list value {default!r} for '{dest}' cannot be coerced to {expected_type.__name__} error: {error}"
                    ) from error

    def _validate_resolver(
        self, action: ArgumentAction, resolver: BaseAction | None
    ) -> BaseAction | None:
        """Validate the action object."""
        if action != ArgumentAction.ACTION and resolver is None:
            return None
        elif action == ArgumentAction.ACTION and resolver is None:
            raise CommandArgumentError("resolver must be provided for ACTION action")
        elif action != ArgumentAction.ACTION and resolver is not None:
            raise CommandArgumentError(
                f"resolver should not be provided for action {action}"
            )

        if not isinstance(resolver, BaseAction):
            raise CommandArgumentError("resolver must be an instance of BaseAction")
        return resolver

    def _validate_action(
        self, action: ArgumentAction | str, positional: bool
    ) -> ArgumentAction:
        if not isinstance(action, ArgumentAction):
            try:
                action = ArgumentAction(action)
            except ValueError:
                raise CommandArgumentError(
                    f"Invalid action '{action}' is not a valid ArgumentAction"
                )
        if action in (
            ArgumentAction.STORE_TRUE,
            ArgumentAction.STORE_FALSE,
            ArgumentAction.STORE_BOOL_OPTIONAL,
            ArgumentAction.COUNT,
            ArgumentAction.HELP,
        ):
            if positional:
                raise CommandArgumentError(
                    f"Action '{action}' cannot be used with positional arguments"
                )

        return action

    def _resolve_default(
        self,
        default: Any,
        action: ArgumentAction,
        nargs: str | int | None,
    ) -> Any:
        """Get the default value for the argument."""
        if default is None:
            if action == ArgumentAction.STORE_TRUE:
                return False
            elif action == ArgumentAction.STORE_FALSE:
                return True
            elif action == ArgumentAction.COUNT:
                return 0
            elif action in (ArgumentAction.APPEND, ArgumentAction.EXTEND):
                return []
            elif isinstance(nargs, int):
                return []
            elif nargs in ("+", "*"):
                return []
            else:
                return None
        elif action in (
            ArgumentAction.STORE_TRUE,
            ArgumentAction.STORE_FALSE,
            ArgumentAction.STORE_BOOL_OPTIONAL,
        ):
            raise CommandArgumentError(
                f"Default value cannot be set for action {action}. It is a boolean flag."
            )
        return default

    def _validate_flags(self, flags: tuple[str, ...]) -> None:
        """Validate the flags provided for the argument."""
        if not flags:
            raise CommandArgumentError("No flags provided")
        for flag in flags:
            if not isinstance(flag, str):
                raise CommandArgumentError(f"Flag '{flag}' must be a string")
            if flag.startswith("--") and len(flag) < 3:
                raise CommandArgumentError(
                    f"Flag '{flag}' must be at least 3 characters long"
                )
            if flag.startswith("-") and not flag.startswith("--") and len(flag) > 2:
                raise CommandArgumentError(
                    f"Flag '{flag}' must be a single character or start with '--'"
                )

    def _register_store_bool_optional(
        self,
        flags: tuple[str, ...],
        dest: str,
        help: str,
    ) -> None:
        if len(flags) != 1:
            raise CommandArgumentError(
                "store_bool_optional action can only have a single flag"
            )
        if not flags[0].startswith("--"):
            raise CommandArgumentError(
                "store_bool_optional action must use a long flag (e.g. --flag)"
            )
        base_flag = flags[0]
        negated_flag = f"--no-{base_flag.lstrip('-')}"

        argument = Argument(
            flags=flags,
            dest=dest,
            action=ArgumentAction.STORE_BOOL_OPTIONAL,
            type=true_none,
            default=None,
            help=help,
        )

        negated_argument = Argument(
            flags=(negated_flag,),
            dest=dest,
            action=ArgumentAction.STORE_BOOL_OPTIONAL,
            type=false_none,
            default=None,
            help=help,
        )

        self._register_argument(argument)
        self._register_argument(negated_argument)

    def _register_argument(self, argument: Argument):

        for flag in argument.flags:
            if (
                flag in self._flag_map
                and not argument.action == ArgumentAction.STORE_BOOL_OPTIONAL
            ):
                existing = self._flag_map[flag]
                raise CommandArgumentError(
                    f"Flag '{flag}' is already used by argument '{existing.dest}'"
                )
        for flag in argument.flags:
            self._flag_map[flag] = argument
            if not argument.positional:
                self._keyword[flag] = argument
        self._dest_set.add(argument.dest)
        self._arguments.append(argument)
        if argument.positional:
            self._positional[argument.dest] = argument
        else:
            self._keyword_list.append(argument)

    def add_argument(
        self,
        *flags,
        action: str | ArgumentAction = "store",
        nargs: int | str | None = None,
        default: Any = None,
        type: Any = str,
        choices: Iterable | None = None,
        required: bool = False,
        help: str = "",
        dest: str | None = None,
        resolver: BaseAction | None = None,
        lazy_resolver: bool = False,
    ) -> None:
        """Add an argument to the parser.
        For `ArgumentAction.ACTION`, `nargs` and `type` determine how many and what kind
        of inputs are passed to the `resolver`.

        The return value of the `resolver` is used directly (no type coercion is applied).
        Validation, structure, and post-processing should be handled within the `resolver`.

        Args:
            name or flags: Either a name or prefixed flags (e.g. 'faylx', '-f', '--falyx').
            action: The action to be taken when the argument is encountered.
            nargs: The number of arguments expected.
            default: The default value if the argument is not provided.
            type: The type to which the command-line argument should be converted.
            choices: A container of the allowable values for the argument.
            required: Whether or not the argument is required.
            help: A brief description of the argument.
            dest: The name of the attribute to be added to the object returned by parse_args().
            resolver: A BaseAction called with optional nargs specified parsed arguments.
        """
        expected_type = type
        self._validate_flags(flags)
        positional = self._is_positional(flags)
        dest = self._get_dest_from_flags(flags, dest)
        if dest in self._dest_set:
            raise CommandArgumentError(
                f"Destination '{dest}' is already defined.\n"
                "Merging multiple arguments into the same dest (e.g. positional + flagged) "
                "is not supported. Define a unique 'dest' for each argument."
            )
        action = self._validate_action(action, positional)
        resolver = self._validate_resolver(action, resolver)

        nargs = self._validate_nargs(nargs, action)
        default = self._resolve_default(default, action, nargs)
        if (
            action in (ArgumentAction.STORE, ArgumentAction.APPEND, ArgumentAction.EXTEND)
            and default is not None
        ):
            if isinstance(default, list):
                self._validate_default_list_type(default, expected_type, dest)
            else:
                self._validate_default_type(default, expected_type, dest)
        choices = self._normalize_choices(choices, expected_type, action)
        if default is not None and choices and default not in choices:
            raise CommandArgumentError(
                f"Default value '{default}' not in allowed choices: {choices}"
            )
        required = self._determine_required(required, positional, nargs, action)
        if not isinstance(lazy_resolver, bool):
            raise CommandArgumentError(
                f"lazy_resolver must be a boolean, got {type(lazy_resolver)}"
            )
        if action == ArgumentAction.STORE_BOOL_OPTIONAL:
            self._register_store_bool_optional(flags, dest, help)
        else:
            argument = Argument(
                flags=flags,
                dest=dest,
                action=action,
                type=expected_type,
                default=default,
                choices=choices,
                required=required,
                help=help,
                nargs=nargs,
                positional=positional,
                resolver=resolver,
                lazy_resolver=lazy_resolver,
            )
            self._register_argument(argument)

    def get_argument(self, dest: str) -> Argument | None:
        return next((a for a in self._arguments if a.dest == dest), None)

    def to_definition_list(self) -> list[dict[str, Any]]:
        defs = []
        for arg in self._arguments:
            defs.append(
                {
                    "flags": arg.flags,
                    "dest": arg.dest,
                    "action": arg.action,
                    "type": arg.type,
                    "choices": arg.choices,
                    "required": arg.required,
                    "nargs": arg.nargs,
                    "positional": arg.positional,
                    "default": arg.default,
                    "help": arg.help,
                }
            )
        return defs

    def _consume_nargs(
        self, args: list[str], start: int, spec: Argument
    ) -> tuple[list[str], int]:
        assert (
            spec.nargs is None
            or isinstance(spec.nargs, int)
            or isinstance(spec.nargs, str)
            and spec.nargs in ("+", "*", "?")
        ), f"Invalid nargs value: {spec.nargs}"
        values = []
        i = start
        if isinstance(spec.nargs, int):
            values = args[i : i + spec.nargs]
            return values, i + spec.nargs
        elif spec.nargs == "+":
            if i >= len(args):
                raise CommandArgumentError(
                    f"Expected at least one value for '{spec.dest}'"
                )
            while i < len(args) and args[i] not in self._keyword:
                values.append(args[i])
                i += 1
            assert values, "Expected at least one value for '+' nargs: shouldn't happen"
            return values, i
        elif spec.nargs == "*":
            while i < len(args) and args[i] not in self._keyword:
                values.append(args[i])
                i += 1
            return values, i
        elif spec.nargs == "?":
            if i < len(args) and args[i] not in self._keyword:
                return [args[i]], i + 1
            return [], i
        elif spec.nargs is None:
            if i < len(args) and args[i] not in self._keyword:
                return [args[i]], i + 1
            return [], i
        assert False, "Invalid nargs value: shouldn't happen"

    async def _consume_all_positional_args(
        self,
        args: list[str],
        result: dict[str, Any],
        positional_args: list[Argument],
        consumed_positional_indicies: set[int],
        from_validate: bool = False,
    ) -> int:
        remaining_positional_args = [
            (j, spec)
            for j, spec in enumerate(positional_args)
            if j not in consumed_positional_indicies
        ]
        i = 0

        for j, spec in remaining_positional_args:
            # estimate how many args the remaining specs might need
            is_last = j == len(positional_args) - 1
            remaining = len(args) - i
            min_required = 0
            for next_spec in positional_args[j + 1 :]:
                assert (
                    next_spec.nargs is None
                    or isinstance(next_spec.nargs, int)
                    or isinstance(next_spec.nargs, str)
                    and next_spec.nargs in ("+", "*", "?")
                ), f"Invalid nargs value: {spec.nargs}"

                if next_spec.default:
                    continue

                if next_spec.nargs is None:
                    min_required += 1
                elif isinstance(next_spec.nargs, int):
                    min_required += next_spec.nargs
                elif next_spec.nargs == "+":
                    min_required += 1
                elif next_spec.nargs == "?":
                    continue
                elif next_spec.nargs == "*":
                    continue

            slice_args = args[i:] if is_last else args[i : i + (remaining - min_required)]
            values, new_i = self._consume_nargs(slice_args, 0, spec)
            i += new_i

            try:
                typed = [coerce_value(value, spec.type) for value in values]
            except Exception as error:
                if len(args[i - new_i :]) == 1 and args[i - new_i].startswith("-"):
                    token = args[i - new_i]
                    valid_flags = [
                        flag for flag in self._flag_map if flag.startswith(token)
                    ]
                    if valid_flags:
                        raise CommandArgumentError(
                            f"Unrecognized option '{token}'. Did you mean one of: {', '.join(valid_flags)}?"
                        ) from error
                    else:
                        raise CommandArgumentError(
                            f"Unrecognized option '{token}'. Use --help to see available options."
                        ) from error
                else:
                    raise CommandArgumentError(
                        f"Invalid value for '{spec.dest}': {error}"
                    ) from error
            if spec.action == ArgumentAction.ACTION:
                assert isinstance(
                    spec.resolver, BaseAction
                ), "resolver should be an instance of BaseAction"
                if not spec.lazy_resolver or not from_validate:
                    try:
                        result[spec.dest] = await spec.resolver(*typed)
                    except Exception as error:
                        raise CommandArgumentError(
                            f"[{spec.dest}] Action failed: {error}"
                        ) from error
            elif not typed and spec.default:
                result[spec.dest] = spec.default
            elif spec.action == ArgumentAction.APPEND:
                assert result.get(spec.dest) is not None, "dest should not be None"
                if spec.nargs is None:
                    result[spec.dest].append(typed[0])
                else:
                    result[spec.dest].append(typed)
            elif spec.action == ArgumentAction.EXTEND:
                assert result.get(spec.dest) is not None, "dest should not be None"
                result[spec.dest].extend(typed)
            elif spec.nargs in (None, 1, "?"):
                result[spec.dest] = typed[0] if len(typed) == 1 else typed
            else:
                result[spec.dest] = typed

            if spec.nargs not in ("*", "+"):
                consumed_positional_indicies.add(j)

        if i < len(args):
            if len(args[i:]) == 1 and args[i].startswith("-"):
                token = args[i]
                valid_flags = [flag for flag in self._flag_map if flag.startswith(token)]
                if valid_flags:
                    raise CommandArgumentError(
                        f"Unrecognized option '{token}'. Did you mean one of: {', '.join(valid_flags)}?"
                    )
                else:
                    raise CommandArgumentError(
                        f"Unrecognized option '{token}'. Use --help to see available options."
                    )
            else:
                plural = "s" if len(args[i:]) > 1 else ""
                raise CommandArgumentError(
                    f"Unexpected positional argument{plural}: {', '.join(args[i:])}"
                )

        return i

    def _expand_posix_bundling(self, token: str) -> list[str] | str:
        """Expand POSIX-style bundled arguments into separate arguments."""
        expanded = []
        if token.startswith("-") and not token.startswith("--") and len(token) > 2:
            # POSIX bundle
            # e.g. -abc -> -a -b -c
            for char in token[1:]:
                flag = f"-{char}"
                arg = self._flag_map.get(flag)
                if not arg:
                    raise CommandArgumentError(f"Unrecognized option: {flag}")
                expanded.append(flag)
        else:
            return token
        return expanded

    async def _handle_token(
        self,
        token: str,
        args: list[str],
        i: int,
        result: dict[str, Any],
        positional_args: list[Argument],
        consumed_positional_indices: set[int],
        consumed_indices: set[int],
        from_validate: bool = False,
    ) -> int:
        if token in self._keyword:
            spec = self._keyword[token]
            action = spec.action

            if action == ArgumentAction.HELP:
                if not from_validate:
                    self.render_help()
                raise HelpSignal()
            elif action == ArgumentAction.ACTION:
                assert isinstance(
                    spec.resolver, BaseAction
                ), "resolver should be an instance of BaseAction"
                values, new_i = self._consume_nargs(args, i + 1, spec)
                try:
                    typed_values = [coerce_value(value, spec.type) for value in values]
                except ValueError as error:
                    raise CommandArgumentError(
                        f"Invalid value for '{spec.dest}': {error}"
                    ) from error
                try:
                    result[spec.dest] = await spec.resolver(*typed_values)
                except Exception as error:
                    raise CommandArgumentError(
                        f"[{spec.dest}] Action failed: {error}"
                    ) from error
                consumed_indices.update(range(i, new_i))
                i = new_i
            elif action == ArgumentAction.STORE_TRUE:
                result[spec.dest] = True
                consumed_indices.add(i)
                i += 1
            elif action == ArgumentAction.STORE_FALSE:
                result[spec.dest] = False
                consumed_indices.add(i)
                i += 1
            elif action == ArgumentAction.STORE_BOOL_OPTIONAL:
                result[spec.dest] = spec.type(True)
                consumed_indices.add(i)
                i += 1
            elif action == ArgumentAction.COUNT:
                result[spec.dest] = result.get(spec.dest, 0) + 1
                consumed_indices.add(i)
                i += 1
            elif action == ArgumentAction.APPEND:
                assert result.get(spec.dest) is not None, "dest should not be None"
                values, new_i = self._consume_nargs(args, i + 1, spec)
                try:
                    typed_values = [coerce_value(value, spec.type) for value in values]
                except ValueError as error:
                    raise CommandArgumentError(
                        f"Invalid value for '{spec.dest}': {error}"
                    ) from error
                if spec.nargs is None:
                    result[spec.dest].append(spec.type(values[0]))
                else:
                    result[spec.dest].append(typed_values)
                consumed_indices.update(range(i, new_i))
                i = new_i
            elif action == ArgumentAction.EXTEND:
                assert result.get(spec.dest) is not None, "dest should not be None"
                values, new_i = self._consume_nargs(args, i + 1, spec)
                try:
                    typed_values = [coerce_value(value, spec.type) for value in values]
                except ValueError as error:
                    raise CommandArgumentError(
                        f"Invalid value for '{spec.dest}': {error}"
                    ) from error
                result[spec.dest].extend(typed_values)
                consumed_indices.update(range(i, new_i))
                i = new_i
            else:
                values, new_i = self._consume_nargs(args, i + 1, spec)
                try:
                    typed_values = [coerce_value(value, spec.type) for value in values]
                except ValueError as error:
                    raise CommandArgumentError(
                        f"Invalid value for '{spec.dest}': {error}"
                    ) from error
                if not typed_values and spec.nargs not in ("*", "?"):
                    choices = []
                    if spec.default:
                        choices.append(f"default={spec.default}")
                    if spec.choices:
                        choices.append(f"choices={spec.choices}")
                    if choices:
                        choices_text = ", ".join(choices)
                        raise CommandArgumentError(
                            f"Argument '{spec.dest}' requires a value. {choices_text}"
                        )
                    elif spec.nargs is None:
                        try:
                            raise CommandArgumentError(
                                f"Enter a {spec.type.__name__} value for '{spec.dest}'"
                            )
                        except AttributeError:
                            raise CommandArgumentError(f"Enter a value for '{spec.dest}'")
                    else:
                        raise CommandArgumentError(
                            f"Argument '{spec.dest}' requires a value. Expected {spec.nargs} values."
                        )
                if spec.nargs in (None, 1, "?") and spec.action != ArgumentAction.APPEND:
                    result[spec.dest] = (
                        typed_values[0] if len(typed_values) == 1 else typed_values
                    )
                else:
                    result[spec.dest] = typed_values
                consumed_indices.update(range(i, new_i))
                i = new_i
        elif token.startswith("-"):
            # Handle unrecognized option
            valid_flags = [flag for flag in self._flag_map if flag.startswith(token)]
            if valid_flags:
                raise CommandArgumentError(
                    f"Unrecognized option '{token}'. Did you mean one of: {', '.join(valid_flags)}?"
                )
            else:
                raise CommandArgumentError(
                    f"Unrecognized option '{token}'. Use --help to see available options."
                )
        else:
            # Get the next flagged argument index if it exists
            next_flagged_index = -1
            for index, arg in enumerate(args[i:], start=i):
                if arg in self._keyword:
                    next_flagged_index = index
                    break
            if next_flagged_index == -1:
                next_flagged_index = len(args)
            args_consumed = await self._consume_all_positional_args(
                args[i:next_flagged_index],
                result,
                positional_args,
                consumed_positional_indices,
                from_validate=from_validate,
            )
            i += args_consumed
        return i

    async def parse_args(
        self, args: list[str] | None = None, from_validate: bool = False
    ) -> dict[str, Any]:
        """Parse Falyx Command arguments."""
        if args is None:
            args = []

        result = {arg.dest: deepcopy(arg.default) for arg in self._arguments}
        positional_args: list[Argument] = [
            arg for arg in self._arguments if arg.positional
        ]
        consumed_positional_indices: set[int] = set()
        consumed_indices: set[int] = set()

        i = 0
        while i < len(args):
            token = self._expand_posix_bundling(args[i])
            if isinstance(token, list):
                args[i : i + 1] = token
                token = args[i]
            i = await self._handle_token(
                token,
                args,
                i,
                result,
                positional_args,
                consumed_positional_indices,
                consumed_indices,
                from_validate=from_validate,
            )

        # Required validation
        for spec in self._arguments:
            if spec.dest == "help":
                continue
            if spec.required and not result.get(spec.dest):
                help_text = f" help: {spec.help}" if spec.help else ""
                if (
                    spec.action == ArgumentAction.ACTION
                    and spec.lazy_resolver
                    and from_validate
                ):
                    continue  # Lazy resolvers are not validated here
                raise CommandArgumentError(
                    f"Missing required argument '{spec.dest}': {spec.get_choice_text()}{help_text}"
                )

            if spec.choices and result.get(spec.dest) not in spec.choices:
                raise CommandArgumentError(
                    f"Invalid value for '{spec.dest}': must be one of {{{', '.join(spec.choices)}}}"
                )

            if spec.action == ArgumentAction.ACTION:
                continue

            if isinstance(spec.nargs, int) and spec.nargs > 1:
                assert isinstance(
                    result.get(spec.dest), list
                ), f"Invalid value for '{spec.dest}': expected a list"
                if not result[spec.dest] and not spec.required:
                    continue
                if spec.action == ArgumentAction.APPEND:
                    for group in result[spec.dest]:
                        if len(group) % spec.nargs != 0:
                            raise CommandArgumentError(
                                f"Invalid number of values for '{spec.dest}': expected a multiple of {spec.nargs}"
                            )
                elif spec.action == ArgumentAction.EXTEND:
                    if len(result[spec.dest]) % spec.nargs != 0:
                        raise CommandArgumentError(
                            f"Invalid number of values for '{spec.dest}': expected a multiple of {spec.nargs}"
                        )
                elif len(result[spec.dest]) != spec.nargs:
                    raise CommandArgumentError(
                        f"Invalid number of values for '{spec.dest}': expected {spec.nargs}, got {len(result[spec.dest])}"
                    )

        result.pop("help", None)
        return result

    async def parse_args_split(
        self, args: list[str], from_validate: bool = False
    ) -> tuple[tuple[Any, ...], dict[str, Any]]:
        """
        Returns:
            tuple[args, kwargs] - Positional arguments in defined order,
            followed by keyword argument mapping.
        """
        parsed = await self.parse_args(args, from_validate)
        args_list = []
        kwargs_dict = {}
        for arg in self._arguments:
            if arg.dest == "help":
                continue
            if arg.positional:
                args_list.append(parsed[arg.dest])
            else:
                kwargs_dict[arg.dest] = parsed[arg.dest]
        return tuple(args_list), kwargs_dict

    def get_options_text(self, plain_text=False) -> str:
        # Options
        # Add all keyword arguments to the options list
        options_list = []
        for arg in self._keyword_list:
            choice_text = arg.get_choice_text()
            if choice_text:
                options_list.extend([f"[{arg.flags[0]} {choice_text}]"])
            else:
                options_list.extend([f"[{arg.flags[0]}]"])

        # Add positional arguments to the options list
        for arg in self._positional.values():
            choice_text = arg.get_choice_text()
            if isinstance(arg.nargs, int):
                choice_text = " ".join([choice_text] * arg.nargs)
            if plain_text:
                options_list.append(choice_text)
            else:
                options_list.append(escape(choice_text))

        return " ".join(options_list)

    def get_command_keys_text(self, plain_text=False) -> str:
        if plain_text:
            command_keys = " | ".join(
                [f"{self.command_key}"] + [f"{alias}" for alias in self.aliases]
            )
        else:
            command_keys = " | ".join(
                [f"[{self.command_style}]{self.command_key}[/{self.command_style}]"]
                + [
                    f"[{self.command_style}]{alias}[/{self.command_style}]"
                    for alias in self.aliases
                ]
            )
        return command_keys

    def get_usage(self, plain_text=False) -> str:
        """Get the usage text for the command."""
        command_keys = self.get_command_keys_text(plain_text)
        options_text = self.get_options_text(plain_text)
        if options_text:
            return f"{command_keys} {options_text}"
        return command_keys

    def render_help(self) -> None:
        usage = self.get_usage()
        self.console.print(f"[bold]usage: {usage}[/bold]\n")

        # Description
        if self.help_text:
            self.console.print(self.help_text + "\n")

        # Arguments
        if self._arguments:
            if self._positional:
                self.console.print("[bold]positional:[/bold]")
                for arg in self._positional.values():
                    flags = arg.get_positional_text()
                    arg_line = f"  {flags:<30} "
                    help_text = arg.help or ""
                    if help_text and len(flags) > 30:
                        help_text = f"\n{'':<33}{help_text}"
                    self.console.print(f"{arg_line}{help_text}")
            self.console.print("[bold]options:[/bold]")
            arg_groups = defaultdict(list)
            for arg in self._keyword_list:
                arg_groups[arg.dest].append(arg)

            for group in arg_groups.values():
                if len(group) == 2 and all(
                    arg.action == ArgumentAction.STORE_BOOL_OPTIONAL for arg in group
                ):
                    # Merge --flag / --no-flag pair into single help line for STORE_BOOL_OPTIONAL
                    all_flags = tuple(
                        sorted(
                            (arg.flags[0] for arg in group),
                            key=lambda f: f.startswith("--no-"),
                        )
                    )
                else:
                    all_flags = group[0].flags

                flags = ", ".join(all_flags)
                flags_choice = f"{flags} {group[0].get_choice_text()}"
                arg_line = f"  {flags_choice:<30} "
                help_text = group[0].help or ""
                if help_text and len(flags_choice) > 30:
                    help_text = f"\n{'':<33}{help_text}"
                self.console.print(f"{arg_line}{help_text}")

        # Epilog
        if self.help_epilog:
            self.console.print("\n" + self.help_epilog, style="dim")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CommandArgumentParser):
            return False

        def sorted_args(parser):
            return sorted(parser._arguments, key=lambda a: a.dest)

        return sorted_args(self) == sorted_args(other)

    def __hash__(self) -> int:
        return hash(tuple(sorted(self._arguments, key=lambda a: a.dest)))

    def __str__(self) -> str:
        positional = sum(arg.positional for arg in self._arguments)
        required = sum(arg.required for arg in self._arguments)
        return (
            f"CommandArgumentParser(args={len(self._arguments)}, "
            f"flags={len(self._flag_map)}, keywords={len(self._keyword)}, "
            f"positional={positional}, required={required})"
        )

    def __repr__(self) -> str:
        return str(self)



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""signals.py"""


class FlowSignal(BaseException):
    """Base class for all flow control signals in Falyx.

    These are not errors. They're used to control flow like quitting,
    going back, or restarting from user input or nested menus.
    """


class BreakChainSignal(FlowSignal):
    """Raised to break the current action chain and return to the previous context."""

    def __init__(self, message: str = "Break chain signal received."):
        super().__init__(message)


class QuitSignal(FlowSignal):
    """Raised to signal an immediate exit from the CLI framework."""

    def __init__(self, message: str = "Quit signal received."):
        super().__init__(message)


class BackSignal(FlowSignal):
    """Raised to return control to the previous menu or caller."""

    def __init__(self, message: str = "Back signal received."):
        super().__init__(message)


class CancelSignal(FlowSignal):
    """Raised to cancel the current command or action."""

    def __init__(self, message: str = "Cancel signal received."):
        super().__init__(message)


class HelpSignal(FlowSignal):
    """Raised to display help information."""

    def __init__(self, message: str = "Help signal received."):
        super().__init__(message)



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""process_action.py"""
from __future__ import annotations

import asyncio
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from typing import Any, Callable

from rich.tree import Tree

from falyx.action.base_action import BaseAction
from falyx.context import ExecutionContext
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import HookManager, HookType
from falyx.themes import OneColors


class ProcessAction(BaseAction):
    """
    ProcessAction runs a function in a separate process using ProcessPoolExecutor.

    Features:
    - Executes CPU-bound or blocking tasks without blocking the main event loop.
    - Supports last_result injection into the subprocess.
    - Validates that last_result is pickleable when injection is enabled.

    Args:
        name (str): Name of the action.
        func (Callable): Function to execute in a new process.
        args (tuple, optional): Positional arguments.
        kwargs (dict, optional): Keyword arguments.
        hooks (HookManager, optional): Hook manager for lifecycle events.
        executor (ProcessPoolExecutor, optional): Custom executor if desired.
        inject_last_result (bool, optional): Inject last result into the function.
        inject_into (str, optional): Name of the injected key.
    """

    def __init__(
        self,
        name: str,
        action: Callable[..., Any],
        *,
        args: tuple = (),
        kwargs: dict[str, Any] | None = None,
        hooks: HookManager | None = None,
        executor: ProcessPoolExecutor | None = None,
        inject_last_result: bool = False,
        inject_into: str = "last_result",
    ):
        super().__init__(
            name,
            hooks=hooks,
            inject_last_result=inject_last_result,
            inject_into=inject_into,
        )
        self.action = action
        self.args = args
        self.kwargs = kwargs or {}
        self.executor = executor or ProcessPoolExecutor()
        self.is_retryable = True

    def get_infer_target(self) -> tuple[Callable[..., Any] | None, None]:
        return self.action, None

    async def _run(self, *args, **kwargs) -> Any:
        if self.inject_last_result and self.shared_context:
            last_result = self.shared_context.last_result()
            if not self._validate_pickleable(last_result):
                raise ValueError(
                    f"Cannot inject last result into {self.name}: "
                    f"last result is not pickleable."
                )
        combined_args = args + self.args
        combined_kwargs = self._maybe_inject_last_result({**self.kwargs, **kwargs})
        context = ExecutionContext(
            name=self.name,
            args=combined_args,
            kwargs=combined_kwargs,
            action=self,
        )
        loop = asyncio.get_running_loop()

        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)
            result = await loop.run_in_executor(
                self.executor, partial(self.action, *combined_args, **combined_kwargs)
            )
            context.result = result
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return result
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            if context.result is not None:
                return context.result
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    def _validate_pickleable(self, obj: Any) -> bool:
        try:
            import pickle

            pickle.dumps(obj)
            return True
        except (pickle.PicklingError, TypeError):
            return False

    async def preview(self, parent: Tree | None = None):
        label = [
            f"[{OneColors.DARK_YELLOW_b}]🧠 ProcessAction (new process)[/] '{self.name}'"
        ]
        if self.inject_last_result:
            label.append(f" [dim](injects '{self.inject_into}')[/dim]")
        if parent:
            parent.add("".join(label))
        else:
            self.console.print(Tree("".join(label)))

    def __str__(self) -> str:
        return (
            f"ProcessAction(name={self.name!r}, "
            f"action={getattr(self.action, '__name__', repr(self.action))}, "
            f"args={self.args!r}, kwargs={self.kwargs!r})"
        )



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""io_action.py
BaseIOAction: A base class for stream- or buffer-based IO-driven Actions.

This module defines `BaseIOAction`, a specialized variant of `BaseAction`
that interacts with standard input and output, enabling command-line pipelines,
text filters, and stream processing tasks.

Features:
- Supports buffered or streaming input modes.
- Reads from stdin and writes to stdout automatically.
- Integrates with lifecycle hooks and retry logic.
- Subclasses must implement `from_input()`, `to_output()`, and `_run()`.

Common usage includes shell-like filters, input transformers, or any tool that
needs to consume input from another process or pipeline.
"""
import asyncio
import sys
from typing import Any, Callable

from rich.tree import Tree

from falyx.action.base_action import BaseAction
from falyx.context import ExecutionContext
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import HookManager, HookType
from falyx.themes import OneColors


class BaseIOAction(BaseAction):
    """
    Base class for IO-driven Actions that operate on stdin/stdout input streams.

    Designed for use in shell pipelines or programmatic workflows that pass data
    through chained commands. It handles reading input, transforming it, and
    emitting output — either as a one-time buffered operation or line-by-line streaming.

    Core responsibilities:
    - Reads input from stdin or previous action result.
    - Supports buffered or streaming modes via `mode`.
    - Parses input via `from_input()` and formats output via `to_output()`.
    - Executes `_run()` with the parsed input.
    - Writes output to stdout.

    Subclasses must implement:
    - `from_input(raw)`: Convert raw stdin or injected data into typed input.
    - `to_output(data)`: Convert result into output string or bytes.
    - `_run(parsed_input, *args, **kwargs)`: Core execution logic.

    Attributes:
        mode (str): Either "buffered" or "stream". Controls input behavior.
        inject_last_result (bool): Whether to inject shared context input.
    """

    def __init__(
        self,
        name: str,
        *,
        hooks: HookManager | None = None,
        mode: str = "buffered",
        logging_hooks: bool = True,
        inject_last_result: bool = True,
    ):
        super().__init__(
            name=name,
            hooks=hooks,
            logging_hooks=logging_hooks,
            inject_last_result=inject_last_result,
        )
        self.mode = mode

    def from_input(self, raw: str | bytes) -> Any:
        raise NotImplementedError

    def to_output(self, result: Any) -> str | bytes:
        raise NotImplementedError

    async def _resolve_input(
        self, args: tuple[Any], kwargs: dict[str, Any]
    ) -> str | bytes:
        data = await self._read_stdin()
        if data:
            return self.from_input(data)

        if len(args) == 1:
            return self.from_input(args[0])

        if self.inject_last_result and self.shared_context:
            return self.shared_context.last_result()

        return ""

    def get_infer_target(self) -> tuple[Callable[..., Any] | None, dict[str, Any] | None]:
        return None, None

    async def __call__(self, *args, **kwargs):
        context = ExecutionContext(
            name=self.name,
            args=args,
            kwargs=kwargs,
            action=self,
        )

        context.start_timer()
        await self.hooks.trigger(HookType.BEFORE, context)

        try:
            if self.mode == "stream":
                line_gen = await self._read_stdin_stream()
                async for _ in self._stream_lines(line_gen, args, kwargs):
                    pass
                result = getattr(self, "_last_result", None)
            else:
                parsed_input = await self._resolve_input(args, kwargs)
                result = await self._run(parsed_input)
                output = self.to_output(result)
                await self._write_stdout(output)
            context.result = result
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return result
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def _read_stdin(self) -> str:
        if not sys.stdin.isatty():
            return await asyncio.to_thread(sys.stdin.read)
        return ""

    async def _read_stdin_stream(self) -> Any:
        """Returns a generator that yields lines from stdin in a background thread."""
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: iter(sys.stdin))

    async def _stream_lines(self, line_gen, args, kwargs):
        for line in line_gen:
            parsed = self.from_input(line)
            result = await self._run(parsed, *args, **kwargs)
            self._last_result = result
            output = self.to_output(result)
            await self._write_stdout(output)
            yield result

    async def _write_stdout(self, data: str) -> None:
        await asyncio.to_thread(sys.stdout.write, data)
        await asyncio.to_thread(sys.stdout.flush)

    async def _run(self, parsed_input: Any, *args, **kwargs) -> Any:
        """Subclasses should override this with actual logic."""
        raise NotImplementedError("Must implement _run()")

    def __str__(self):
        return f"<{self.__class__.__name__} '{self.name}' IOAction>"

    async def preview(self, parent: Tree | None = None):
        label = [f"[{OneColors.GREEN_b}]⚙ IOAction[/] '{self.name}'"]
        if self.inject_last_result:
            label.append(f" [dim](injects '{self.inject_into}')[/dim]")
        if parent:
            parent.add("".join(label))
        else:
            self.console.print(Tree("".join(label)))



from __future__ import annotations

from enum import Enum
from typing import Any

from prompt_toolkit import PromptSession
from rich.tree import Tree

from falyx.action.base_action import BaseAction
from falyx.context import ExecutionContext
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import HookType
from falyx.logger import logger
from falyx.prompt_utils import confirm_async, should_prompt_user
from falyx.signals import CancelSignal
from falyx.themes import OneColors
from falyx.validators import word_validator, words_validator


class ConfirmType(Enum):
    """Enum for different confirmation types."""

    YES_NO = "yes_no"
    YES_CANCEL = "yes_cancel"
    YES_NO_CANCEL = "yes_no_cancel"
    TYPE_WORD = "type_word"
    OK_CANCEL = "ok_cancel"

    @classmethod
    def choices(cls) -> list[ConfirmType]:
        """Return a list of all hook type choices."""
        return list(cls)

    def __str__(self) -> str:
        """Return the string representation of the confirm type."""
        return self.value


class ConfirmAction(BaseAction):
    """
    Action to confirm an operation with the user.

    There are several ways to confirm an action, such as using a simple
    yes/no prompt. You can also use a confirmation type that requires the user
    to type a specific word or phrase to confirm the action, or use an OK/Cancel
    dialog.

    This action can be used to ensure that the user explicitly agrees to proceed
    with an operation.

    Attributes:
        name (str): Name of the action.
        message (str): The confirmation message to display.
        confirm_type (ConfirmType | str): The type of confirmation to use.
            Options include YES_NO, YES_CANCEL, YES_NO_CANCEL, TYPE_WORD, and OK_CANCEL.
        prompt_session (PromptSession | None): The session to use for input.
        confirm (bool): Whether to prompt the user for confirmation.
        word (str): The word to type for TYPE_WORD confirmation.
        return_last_result (bool): Whether to return the last result of the action
                                   instead of a boolean.
    """

    def __init__(
        self,
        name: str,
        message: str = "Confirm?",
        confirm_type: ConfirmType | str = ConfirmType.YES_NO,
        prompt_session: PromptSession | None = None,
        confirm: bool = True,
        word: str = "CONFIRM",
        return_last_result: bool = False,
        inject_last_result: bool = True,
        inject_into: str = "last_result",
    ):
        """
        Initialize the ConfirmAction.

        Args:
            message (str): The confirmation message to display.
            confirm_type (ConfirmType): The type of confirmation to use.
                Options include YES_NO, YES_CANCEL, YES_NO_CANCEL, TYPE_WORD, and OK_CANCEL.
            prompt_session (PromptSession | None): The session to use for input.
            confirm (bool): Whether to prompt the user for confirmation.
            word (str): The word to type for TYPE_WORD confirmation.
            return_last_result (bool): Whether to return the last result of the action.
        """
        super().__init__(
            name=name,
            inject_last_result=inject_last_result,
            inject_into=inject_into,
        )
        self.message = message
        self.confirm_type = self._coerce_confirm_type(confirm_type)
        self.prompt_session = prompt_session or PromptSession()
        self.confirm = confirm
        self.word = word
        self.return_last_result = return_last_result

    def _coerce_confirm_type(self, confirm_type: ConfirmType | str) -> ConfirmType:
        """Coerce the confirm_type to a ConfirmType enum."""
        if isinstance(confirm_type, ConfirmType):
            return confirm_type
        elif isinstance(confirm_type, str):
            return ConfirmType(confirm_type)
        return ConfirmType(confirm_type)

    async def _confirm(self) -> bool:
        """Confirm the action with the user."""
        match self.confirm_type:
            case ConfirmType.YES_NO:
                return await confirm_async(
                    self.message,
                    prefix="❓ ",
                    suffix=" [Y/n] > ",
                    session=self.prompt_session,
                )
            case ConfirmType.YES_NO_CANCEL:
                error_message = "Enter 'Y', 'y' to confirm, 'N', 'n' to decline, or 'C', 'c' to abort."
                answer = await self.prompt_session.prompt_async(
                    f"❓ {self.message} [Y]es, [N]o, or [C]ancel to abort > ",
                    validator=words_validator(
                        ["Y", "N", "C"], error_message=error_message
                    ),
                )
                if answer.upper() == "C":
                    raise CancelSignal(f"Action '{self.name}' was cancelled by the user.")
                return answer.upper() == "Y"
            case ConfirmType.TYPE_WORD:
                answer = await self.prompt_session.prompt_async(
                    f"❓ {self.message} [{self.word}] to confirm or [N/n] > ",
                    validator=word_validator(self.word),
                )
                return answer.upper().strip() != "N"
            case ConfirmType.YES_CANCEL:
                answer = await confirm_async(
                    self.message,
                    prefix="❓ ",
                    suffix=" [Y/n] > ",
                    session=self.prompt_session,
                )
                if not answer:
                    raise CancelSignal(f"Action '{self.name}' was cancelled by the user.")
                return answer
            case ConfirmType.OK_CANCEL:
                error_message = "Enter 'O', 'o' to confirm or 'C', 'c' to abort."
                answer = await self.prompt_session.prompt_async(
                    f"❓ {self.message} [O]k to confirm, [C]ancel to abort > ",
                    validator=words_validator(["O", "C"], error_message=error_message),
                )
                if answer.upper() == "C":
                    raise CancelSignal(f"Action '{self.name}' was cancelled by the user.")
                return answer.upper() == "O"
            case _:
                raise ValueError(f"Unknown confirm_type: {self.confirm_type}")

    def get_infer_target(self) -> tuple[None, None]:
        return None, None

    async def _run(self, *args, **kwargs) -> Any:
        combined_kwargs = self._maybe_inject_last_result(kwargs)
        context = ExecutionContext(
            name=self.name, args=args, kwargs=combined_kwargs, action=self
        )
        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)
            if (
                not self.confirm
                or self.options_manager
                and not should_prompt_user(
                    confirm=self.confirm, options=self.options_manager
                )
            ):
                logger.debug(
                    "Skipping confirmation for action '%s' as 'confirm' is False or options manager indicates no prompt.",
                    self.name,
                )
                if self.return_last_result:
                    result = combined_kwargs[self.inject_into]
                else:
                    result = True
            else:
                answer = await self._confirm()
                if self.return_last_result and answer:
                    result = combined_kwargs[self.inject_into]
                else:
                    result = answer
            logger.debug("Action '%s' confirmed with result: %s", self.name, result)
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return result
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def preview(self, parent: Tree | None = None) -> None:
        tree = (
            Tree(
                f"[{OneColors.CYAN_b}]ConfirmAction[/]: {self.name}",
                guide_style=OneColors.BLUE_b,
            )
            if not parent
            else parent.add(f"[{OneColors.CYAN_b}]ConfirmAction[/]: {self.name}")
        )
        tree.add(f"[bold]Message:[/] {self.message}")
        tree.add(f"[bold]Type:[/] {self.confirm_type.value}")
        tree.add(f"[bold]Prompt Required:[/] {'Yes' if self.confirm else 'No'}")
        if self.confirm_type == ConfirmType.TYPE_WORD:
            tree.add(f"[bold]Confirmation Word:[/] {self.word}")
        if parent is None:
            self.console.print(tree)

    def __str__(self) -> str:
        return (
            f"ConfirmAction(name={self.name}, message={self.message}, "
            f"confirm_type={self.confirm_type}, return_last_result={self.return_last_result})"
        )



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""signal_action.py"""
from rich.tree import Tree

from falyx.action.action import Action
from falyx.signals import FlowSignal
from falyx.themes import OneColors


class SignalAction(Action):
    """
    An action that raises a control flow signal when executed.

    Useful for exiting a menu, going back, or halting execution gracefully.
    """

    def __init__(self, name: str, signal: FlowSignal):
        self.signal = signal
        super().__init__(name, action=self.raise_signal)

    async def raise_signal(self, *args, **kwargs):
        raise self.signal

    @property
    def signal(self):
        return self._signal

    @signal.setter
    def signal(self, value: FlowSignal):
        if not isinstance(value, FlowSignal):
            raise TypeError(
                f"Signal must be an FlowSignal instance, got {type(value).__name__}"
            )
        self._signal = value

    def __str__(self):
        return f"SignalAction(name={self.name}, signal={self._signal.__class__.__name__})"

    async def preview(self, parent: Tree | None = None):
        label = f"[{OneColors.LIGHT_RED}]⚡ SignalAction[/] '{self.signal.__class__.__name__}'"
        tree = parent.add(label) if parent else Tree(label)
        if not parent:
            self.console.print(tree)



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""exceptions.py"""


class FalyxError(Exception):
    """Custom exception for the Menu class."""


class CommandAlreadyExistsError(FalyxError):
    """Exception raised when an command with the same key already exists in the menu."""


class InvalidHookError(FalyxError):
    """Exception raised when a hook is not callable."""


class InvalidActionError(FalyxError):
    """Exception raised when an action is not callable."""


class NotAFalyxError(FalyxError):
    """Exception raised when the provided submenu is not an instance of Menu."""


class CircuitBreakerOpen(FalyxError):
    """Exception raised when the circuit breaker is open."""


class EmptyChainError(FalyxError):
    """Exception raised when the chain is empty."""


class EmptyGroupError(FalyxError):
    """Exception raised when the chain is empty."""


class EmptyPoolError(FalyxError):
    """Exception raised when the chain is empty."""


class CommandArgumentError(FalyxError):
    """Exception raised when there is an error in the command argument parser."""



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""chained_action.py"""
from __future__ import annotations

from typing import Any, Awaitable, Callable, Sequence

from rich.tree import Tree

from falyx.action.action import Action
from falyx.action.action_mixins import ActionListMixin
from falyx.action.base_action import BaseAction
from falyx.action.fallback_action import FallbackAction
from falyx.action.literal_input_action import LiteralInputAction
from falyx.context import ExecutionContext, SharedContext
from falyx.exceptions import EmptyChainError
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import Hook, HookManager, HookType
from falyx.logger import logger
from falyx.options_manager import OptionsManager
from falyx.signals import BreakChainSignal
from falyx.themes import OneColors


class ChainedAction(BaseAction, ActionListMixin):
    """
    ChainedAction executes a sequence of actions one after another.

    Features:
    - Supports optional automatic last_result injection (auto_inject).
    - Recovers from intermediate errors using FallbackAction if present.
    - Rolls back all previously executed actions if a failure occurs.
    - Handles literal values with LiteralInputAction.

    Best used for defining robust, ordered workflows where each step can depend on
    previous results.

    Args:
        name (str): Name of the chain.
        actions (list): List of actions or literals to execute.
        hooks (HookManager, optional): Hooks for lifecycle events.
        inject_last_result (bool, optional): Whether to inject last results into kwargs
                                             by default.
        inject_into (str, optional): Key name for injection.
        auto_inject (bool, optional): Auto-enable injection for subsequent actions.
        return_list (bool, optional): Whether to return a list of all results. False
                                      returns the last result.
    """

    def __init__(
        self,
        name: str,
        actions: (
            Sequence[BaseAction | Callable[..., Any] | Callable[..., Awaitable[Any]]]
            | None
        ) = None,
        *,
        hooks: HookManager | None = None,
        inject_last_result: bool = False,
        inject_into: str = "last_result",
        auto_inject: bool = False,
        return_list: bool = False,
    ) -> None:
        super().__init__(
            name,
            hooks=hooks,
            inject_last_result=inject_last_result,
            inject_into=inject_into,
        )
        ActionListMixin.__init__(self)
        self.auto_inject = auto_inject
        self.return_list = return_list
        if actions:
            self.set_actions(actions)

    def _wrap_if_needed(self, action: BaseAction | Callable[..., Any]) -> BaseAction:
        if isinstance(action, BaseAction):
            return action
        elif callable(action):
            return Action(name=action.__name__, action=action)
        else:
            return LiteralInputAction(action)

    def add_action(self, action: BaseAction | Callable[..., Any]) -> None:
        action = self._wrap_if_needed(action)
        if self.actions and self.auto_inject and not action.inject_last_result:
            action.inject_last_result = True
        super().add_action(action)
        if hasattr(action, "register_teardown") and callable(action.register_teardown):
            action.register_teardown(self.hooks)

    def set_actions(self, actions: Sequence[BaseAction | Callable[..., Any]]) -> None:
        """Replaces the current action list with a new one."""
        self.actions.clear()
        for action in actions:
            self.add_action(action)

    def set_options_manager(self, options_manager: OptionsManager) -> None:
        super().set_options_manager(options_manager)
        for action in self.actions:
            action.set_options_manager(options_manager)

    def get_infer_target(self) -> tuple[Callable[..., Any] | None, dict[str, Any] | None]:
        if self.actions:
            return self.actions[0].get_infer_target()
        return None, None

    def _clear_args(self):
        return (), {}

    async def _run(self, *args, **kwargs) -> Any:
        if not self.actions:
            raise EmptyChainError(f"[{self.name}] No actions to execute.")

        shared_context = SharedContext(name=self.name, action=self)
        if self.shared_context:
            shared_context.add_result(self.shared_context.last_result())
        updated_kwargs = self._maybe_inject_last_result(kwargs)
        context = ExecutionContext(
            name=self.name,
            args=args,
            kwargs=updated_kwargs,
            action=self,
            extra={"results": [], "rollback_stack": []},
            shared_context=shared_context,
        )
        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)

            for index, action in enumerate(self.actions):
                if action._skip_in_chain:
                    logger.debug(
                        "[%s] Skipping consumed action '%s'", self.name, action.name
                    )
                    continue
                shared_context.current_index = index
                prepared = action.prepare(shared_context, self.options_manager)
                try:
                    result = await prepared(*args, **updated_kwargs)
                except Exception as error:
                    if index + 1 < len(self.actions) and isinstance(
                        self.actions[index + 1], FallbackAction
                    ):
                        logger.warning(
                            "[%s] Fallback triggered: %s, recovering with fallback "
                            "'%s'.",
                            self.name,
                            error,
                            self.actions[index + 1].name,
                        )
                        shared_context.add_result(None)
                        context.extra["results"].append(None)
                        fallback = self.actions[index + 1].prepare(shared_context)
                        result = await fallback()
                        fallback._skip_in_chain = True
                    else:
                        raise
                args, updated_kwargs = self._clear_args()
                shared_context.add_result(result)
                context.extra["results"].append(result)
                context.extra["rollback_stack"].append(prepared)

            all_results = context.extra["results"]
            assert (
                all_results
            ), f"[{self.name}] No results captured. Something seriously went wrong."
            context.result = all_results if self.return_list else all_results[-1]
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return context.result
        except BreakChainSignal as error:
            logger.info("[%s] Chain broken: %s", self.name, error)
            context.exception = error
            shared_context.add_error(shared_context.current_index, error)
            await self._rollback(context.extra["rollback_stack"], *args, **kwargs)
        except Exception as error:
            context.exception = error
            shared_context.add_error(shared_context.current_index, error)
            await self._rollback(context.extra["rollback_stack"], *args, **kwargs)
            await self.hooks.trigger(HookType.ON_ERROR, context)
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def _rollback(self, rollback_stack, *args, **kwargs):
        """
        Roll back all executed actions in reverse order.

        Rollbacks run even if a fallback recovered from failure,
        ensuring consistent undo of all side effects.

        Actions without rollback handlers are skipped.

        Args:
            rollback_stack (list): Actions to roll back.
            *args, **kwargs: Passed to rollback handlers.
        """
        for action in reversed(rollback_stack):
            rollback = getattr(action, "rollback", None)
            if rollback:
                try:
                    logger.warning("[%s] Rolling back...", action.name)
                    await action.rollback(*args, **kwargs)
                except Exception as error:
                    logger.error("[%s] Rollback failed: %s", action.name, error)

    def register_hooks_recursively(self, hook_type: HookType, hook: Hook):
        """Register a hook for all actions and sub-actions."""
        super().register_hooks_recursively(hook_type, hook)
        for action in self.actions:
            action.register_hooks_recursively(hook_type, hook)

    async def preview(self, parent: Tree | None = None):
        label = [f"[{OneColors.CYAN_b}]⛓ ChainedAction[/] '{self.name}'"]
        if self.inject_last_result:
            label.append(f" [dim](injects '{self.inject_into}')[/dim]")
        tree = parent.add("".join(label)) if parent else Tree("".join(label))
        for action in self.actions:
            await action.preview(parent=tree)
        if not parent:
            self.console.print(tree)

    def __str__(self):
        return (
            f"ChainedAction(name={self.name!r}, "
            f"actions={[a.name for a in self.actions]!r}, "
            f"auto_inject={self.auto_inject}, return_list={self.return_list})"
        )



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""prompt_menu_action.py"""
from typing import Any

from prompt_toolkit import PromptSession
from prompt_toolkit.formatted_text import FormattedText, merge_formatted_text
from rich.tree import Tree

from falyx.action.base_action import BaseAction
from falyx.context import ExecutionContext
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import HookType
from falyx.logger import logger
from falyx.menu import MenuOptionMap
from falyx.signals import BackSignal, QuitSignal
from falyx.themes import OneColors


class PromptMenuAction(BaseAction):
    """PromptMenuAction class for creating prompt -> actions."""

    def __init__(
        self,
        name: str,
        menu_options: MenuOptionMap,
        *,
        prompt_message: str = "Select > ",
        default_selection: str = "",
        inject_last_result: bool = False,
        inject_into: str = "last_result",
        prompt_session: PromptSession | None = None,
        never_prompt: bool = False,
        include_reserved: bool = True,
    ):
        super().__init__(
            name,
            inject_last_result=inject_last_result,
            inject_into=inject_into,
            never_prompt=never_prompt,
        )
        self.menu_options = menu_options
        self.prompt_message = prompt_message
        self.default_selection = default_selection
        self.prompt_session = prompt_session or PromptSession()
        self.include_reserved = include_reserved

    def get_infer_target(self) -> tuple[None, None]:
        return None, None

    async def _run(self, *args, **kwargs) -> Any:
        kwargs = self._maybe_inject_last_result(kwargs)
        context = ExecutionContext(
            name=self.name,
            args=args,
            kwargs=kwargs,
            action=self,
        )

        effective_default = self.default_selection
        maybe_result = str(self.last_result)
        if maybe_result in self.menu_options:
            effective_default = maybe_result
        elif self.inject_last_result:
            logger.warning(
                "[%s] Injected last result '%s' not found in menu options",
                self.name,
                maybe_result,
            )

        if self.never_prompt and not effective_default:
            raise ValueError(
                f"[{self.name}] 'never_prompt' is True but no valid default_selection"
                " was provided."
            )

        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)
            key = effective_default
            if not self.never_prompt:
                placeholder_formatted_text = []
                for index, (key, option) in enumerate(self.menu_options.items()):
                    placeholder_formatted_text.append(option.render_prompt(key))
                    if index < len(self.menu_options) - 1:
                        placeholder_formatted_text.append(
                            FormattedText([(OneColors.WHITE, " | ")])
                        )
                placeholder = merge_formatted_text(placeholder_formatted_text)
                key = await self.prompt_session.prompt_async(
                    message=self.prompt_message, placeholder=placeholder
                )
            option = self.menu_options[key]
            result = await option.action(*args, **kwargs)
            context.result = result
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return result

        except BackSignal:
            logger.debug("[%s][BackSignal] ← Returning to previous menu", self.name)
            return None
        except QuitSignal:
            logger.debug("[%s][QuitSignal] ← Exiting application", self.name)
            raise
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def preview(self, parent: Tree | None = None):
        label = f"[{OneColors.LIGHT_YELLOW_b}]📋 PromptMenuAction[/] '{self.name}'"
        tree = parent.add(label) if parent else Tree(label)
        for key, option in self.menu_options.items():
            tree.add(
                f"[dim]{key}[/]: {option.description} → [italic]{option.action.name}[/]"
            )
            await option.action.preview(parent=tree)
        if not parent:
            self.console.print(tree)

    def __str__(self) -> str:
        return (
            f"PromptMenuAction(name={self.name!r}, options={list(self.menu_options.keys())!r}, "
            f"default_selection={self.default_selection!r}, "
            f"include_reserved={self.include_reserved}, "
            f"prompt={'off' if self.never_prompt else 'on'})"
        )



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""argument.py"""
from dataclasses import dataclass
from typing import Any

from falyx.action.base_action import BaseAction
from falyx.parser.argument_action import ArgumentAction


@dataclass
class Argument:
    """
    Represents a command-line argument.

    Attributes:
        flags (tuple[str, ...]): Short and long flags for the argument.
        dest (str): The destination name for the argument.
        action (ArgumentAction): The action to be taken when the argument is encountered.
        type (Any): The type of the argument (e.g., str, int, float) or a callable that converts the argument value.
        default (Any): The default value if the argument is not provided.
        choices (list[str] | None): A list of valid choices for the argument.
        required (bool): True if the argument is required, False otherwise.
        help (str): Help text for the argument.
        nargs (int | str | None): Number of arguments expected. Can be an int, '?', '*', '+', or None.
        positional (bool): True if the argument is positional (no leading - or -- in flags), False otherwise.
        resolver (BaseAction | None):
            An action object that resolves the argument, if applicable.
        lazy_resolver (bool): True if the resolver should be called lazily, False otherwise
    """

    flags: tuple[str, ...]
    dest: str
    action: ArgumentAction = ArgumentAction.STORE
    type: Any = str
    default: Any = None
    choices: list[str] | None = None
    required: bool = False
    help: str = ""
    nargs: int | str | None = None
    positional: bool = False
    resolver: BaseAction | None = None
    lazy_resolver: bool = False

    def get_positional_text(self) -> str:
        """Get the positional text for the argument."""
        text = ""
        if self.positional:
            if self.choices:
                text = f"{{{','.join([str(choice) for choice in self.choices])}}}"
            else:
                text = self.dest
        return text

    def get_choice_text(self) -> str:
        """Get the choice text for the argument."""
        choice_text = ""
        if self.choices:
            choice_text = f"{{{','.join([str(choice) for choice in self.choices])}}}"
        elif (
            self.action
            in (
                ArgumentAction.STORE,
                ArgumentAction.APPEND,
                ArgumentAction.EXTEND,
                ArgumentAction.ACTION,
            )
            and not self.positional
        ):
            choice_text = self.dest.upper()
        elif self.action in (
            ArgumentAction.STORE,
            ArgumentAction.APPEND,
            ArgumentAction.EXTEND,
            ArgumentAction.ACTION,
        ) or isinstance(self.nargs, str):
            choice_text = self.dest

        if self.nargs == "?":
            choice_text = f"[{choice_text}]"
        elif self.nargs == "*":
            choice_text = f"[{choice_text} ...]"
        elif self.nargs == "+":
            choice_text = f"{choice_text} [{choice_text} ...]"
        return choice_text

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Argument):
            return False
        return (
            self.flags == other.flags
            and self.dest == other.dest
            and self.action == other.action
            and self.type == other.type
            and self.choices == other.choices
            and self.required == other.required
            and self.nargs == other.nargs
            and self.positional == other.positional
            and self.default == other.default
            and self.help == other.help
        )

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self.flags),
                self.dest,
                self.action,
                self.type,
                tuple(self.choices or []),
                self.required,
                self.nargs,
                self.positional,
                self.default,
                self.help,
            )
        )



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""debug.py"""
from falyx.context import ExecutionContext
from falyx.hook_manager import HookManager, HookType
from falyx.logger import logger


def log_before(context: ExecutionContext):
    """Log the start of an action."""
    args = ", ".join(map(repr, context.args))
    kwargs = ", ".join(f"{key}={value!r}" for key, value in context.kwargs.items())
    signature = ", ".join(filter(None, [args, kwargs]))
    logger.info("[%s] Starting -> %s(%s)", context.name, context.action, signature)


def log_success(context: ExecutionContext):
    """Log the successful completion of an action."""
    result_str = repr(context.result)
    if len(result_str) > 100:
        result_str = f"{result_str[:100]} ..."
    logger.debug("[%s] Success -> Result: %s", context.name, result_str)


def log_after(context: ExecutionContext):
    """Log the completion of an action, regardless of success or failure."""
    logger.debug("[%s] Finished in %.3fs", context.name, context.duration)


def log_error(context: ExecutionContext):
    """Log an error that occurred during the action."""
    logger.error(
        "[%s] Error (%s): %s",
        context.name,
        type(context.exception).__name__,
        context.exception,
        exc_info=True,
    )


def register_debug_hooks(hooks: HookManager):
    hooks.register(HookType.BEFORE, log_before)
    hooks.register(HookType.AFTER, log_after)
    hooks.register(HookType.ON_SUCCESS, log_success)
    hooks.register(HookType.ON_ERROR, log_error)



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""init.py"""
from pathlib import Path

from falyx.console import console

TEMPLATE_TASKS = """\
# This file is used by falyx.yaml to define CLI actions.
# You can run: falyx run [key] or falyx list to see available commands.

import asyncio
import json

from falyx.action import Action, ChainedAction, ShellAction, SelectionAction


post_ids = ["1", "2", "3", "4", "5"]

pick_post = SelectionAction(
    name="Pick Post ID",
    selections=post_ids,
    title="Choose a Post ID",
    prompt_message="Select a post > ",
)

fetch_post = ShellAction(
    name="Fetch Post via curl",
    command_template="curl https://jsonplaceholder.typicode.com/posts/{}",
)

async def get_post_title(last_result):
    return json.loads(last_result).get("title", "No title found")

post_flow = ChainedAction(
    name="Fetch and Parse Post",
    actions=[pick_post, fetch_post, get_post_title],
    auto_inject=True,
)

async def hello():
    print("👋 Hello from Falyx!")
    return "Hello Complete!"

async def some_work():
    await asyncio.sleep(2)
    print("Work Finished!")
    return "Work Complete!"

work_action = Action(
    name="Work Action",
    action=some_work,
)
"""

TEMPLATE_CONFIG = """\
# falyx.yaml — Config-driven CLI definition
# Define your commands here and point to Python callables in tasks.py
title: Sample CLI Project
prompt:
  - ["#61AFEF bold", "FALYX > "]
columns: 3
welcome_message: "🚀 Welcome to your new CLI project!"
exit_message: "👋 See you next time!"
commands:
  - key: S
    description: Say Hello
    action: tasks.hello
    aliases: [hi, hello]
    tags: [example]

  - key: P
    description: Get Post Title
    action: tasks.post_flow
    aliases: [submit]
    preview_before_confirm: true
    confirm: true
    tags: [demo, network]

  - key: G
    description: Do Some Work
    action: tasks.work_action
    aliases: [work]
    spinner: true
    spinner_message: "Working..."
"""

GLOBAL_TEMPLATE_TASKS = """\
async def cleanup():
    print("🧹 Cleaning temp files...")
"""

GLOBAL_CONFIG = """\
title: Global Falyx Config
commands:
  - key: C
    description: Cleanup temp files
    action: tasks.cleanup
    aliases: [clean, cleanup]
"""


def init_project(name: str) -> None:
    target = Path(name).resolve()
    target.mkdir(parents=True, exist_ok=True)

    tasks_path = target / "tasks.py"
    config_path = target / "falyx.yaml"

    if tasks_path.exists() or config_path.exists():
        console.print(f"⚠️  Project already initialized at {target}")
        return None

    tasks_path.write_text(TEMPLATE_TASKS)
    config_path.write_text(TEMPLATE_CONFIG)

    console.print(f"✅ Initialized Falyx project in {target}")


def init_global() -> None:
    config_dir = Path.home() / ".config" / "falyx"
    config_dir.mkdir(parents=True, exist_ok=True)

    tasks_path = config_dir / "tasks.py"
    config_path = config_dir / "falyx.yaml"

    if tasks_path.exists() or config_path.exists():
        console.print("⚠️  Global Falyx config already exists at ~/.config/falyx")
        return None

    tasks_path.write_text(GLOBAL_TEMPLATE_TASKS)
    config_path.write_text(GLOBAL_CONFIG)

    console.print("✅ Initialized global Falyx config at ~/.config/falyx")



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""load_file_action.py"""
import csv
import json
import xml.etree.ElementTree as ET
from datetime import datetime
from pathlib import Path
from typing import Any

import toml
import yaml
from rich.tree import Tree

from falyx.action.action_types import FileType
from falyx.action.base_action import BaseAction
from falyx.context import ExecutionContext
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import HookType
from falyx.logger import logger
from falyx.themes import OneColors


class LoadFileAction(BaseAction):
    """LoadFileAction allows loading and parsing files of various types."""

    def __init__(
        self,
        name: str,
        file_path: str | Path | None = None,
        file_type: FileType | str = FileType.TEXT,
        inject_last_result: bool = False,
        inject_into: str = "file_path",
    ):
        super().__init__(
            name=name, inject_last_result=inject_last_result, inject_into=inject_into
        )
        self._file_path = self._coerce_file_path(file_path)
        self._file_type = self._coerce_file_type(file_type)

    @property
    def file_path(self) -> Path | None:
        """Get the file path as a Path object."""
        return self._file_path

    @file_path.setter
    def file_path(self, value: str | Path):
        """Set the file path, converting to Path if necessary."""
        self._file_path = self._coerce_file_path(value)

    def _coerce_file_path(self, file_path: str | Path | None) -> Path | None:
        """Coerce the file path to a Path object."""
        if isinstance(file_path, Path):
            return file_path
        elif isinstance(file_path, str):
            return Path(file_path)
        elif file_path is None:
            return None
        else:
            raise TypeError("file_path must be a string or Path object")

    @property
    def file_type(self) -> FileType:
        """Get the file type."""
        return self._file_type

    @file_type.setter
    def file_type(self, value: FileType | str):
        """Set the file type, converting to FileType if necessary."""
        self._file_type = self._coerce_file_type(value)

    def _coerce_file_type(self, file_type: FileType | str) -> FileType:
        """Coerce the file type to a FileType enum."""
        if isinstance(file_type, FileType):
            return file_type
        elif isinstance(file_type, str):
            return FileType(file_type)
        else:
            raise TypeError("file_type must be a FileType enum or string")

    def get_infer_target(self) -> tuple[None, None]:
        return None, None

    async def load_file(self) -> Any:
        """Load and parse the file based on its type."""
        if self.file_path is None:
            raise ValueError("file_path must be set before loading a file")
        elif not self.file_path.exists():
            raise FileNotFoundError(f"File not found: {self.file_path}")
        elif not self.file_path.is_file():
            raise ValueError(f"Path is not a regular file: {self.file_path}")
        value: Any = None
        try:
            if self.file_type == FileType.TEXT:
                value = self.file_path.read_text(encoding="UTF-8")
            elif self.file_type == FileType.PATH:
                value = self.file_path
            elif self.file_type == FileType.JSON:
                value = json.loads(self.file_path.read_text(encoding="UTF-8"))
            elif self.file_type == FileType.TOML:
                value = toml.loads(self.file_path.read_text(encoding="UTF-8"))
            elif self.file_type == FileType.YAML:
                value = yaml.safe_load(self.file_path.read_text(encoding="UTF-8"))
            elif self.file_type == FileType.CSV:
                with open(self.file_path, newline="", encoding="UTF-8") as csvfile:
                    reader = csv.reader(csvfile)
                    value = list(reader)
            elif self.file_type == FileType.TSV:
                with open(self.file_path, newline="", encoding="UTF-8") as tsvfile:
                    reader = csv.reader(tsvfile, delimiter="\t")
                    value = list(reader)
            elif self.file_type == FileType.XML:
                tree = ET.parse(self.file_path, parser=ET.XMLParser(encoding="UTF-8"))
                root = tree.getroot()
                value = ET.tostring(root, encoding="unicode")
            else:
                raise ValueError(f"Unsupported return type: {self.file_type}")

        except Exception as error:
            logger.error("Failed to parse %s: %s", self.file_path.name, error)
        return value

    async def _run(self, *args, **kwargs) -> Any:
        context = ExecutionContext(name=self.name, args=args, kwargs=kwargs, action=self)
        context.start_timer()
        try:
            await self.hooks.trigger(HookType.BEFORE, context)

            if "file_path" in kwargs:
                self.file_path = kwargs["file_path"]
            elif self.inject_last_result and self.last_result:
                self.file_path = self.last_result

            result = await self.load_file()
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return result
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def preview(self, parent: Tree | None = None):
        label = f"[{OneColors.GREEN}]📄 LoadFileAction[/] '{self.name}'"
        tree = parent.add(label) if parent else Tree(label)

        tree.add(f"[dim]Path:[/] {self.file_path}")
        tree.add(f"[dim]Type:[/] {self.file_type.name if self.file_type else 'None'}")
        if self.file_path is None:
            tree.add(f"[{OneColors.DARK_RED_b}]❌ File path is not set[/]")
        elif not self.file_path.exists():
            tree.add(f"[{OneColors.DARK_RED_b}]❌ File does not exist[/]")
        elif not self.file_path.is_file():
            tree.add(f"[{OneColors.LIGHT_YELLOW_b}]⚠️ Not a regular file[/]")
        else:
            try:
                stat = self.file_path.stat()
                tree.add(f"[dim]Size:[/] {stat.st_size:,} bytes")
                tree.add(
                    f"[dim]Modified:[/] {datetime.fromtimestamp(stat.st_mtime):%Y-%m-%d %H:%M:%S}"
                )
                tree.add(
                    f"[dim]Created:[/] {datetime.fromtimestamp(stat.st_ctime):%Y-%m-%d %H:%M:%S}"
                )
                if self.file_type == FileType.TEXT:
                    preview_lines = self.file_path.read_text(
                        encoding="UTF-8"
                    ).splitlines()[:10]
                    content_tree = tree.add("[dim]Preview (first 10 lines):[/]")
                    for line in preview_lines:
                        content_tree.add(f"[dim]{line}[/]")
                elif self.file_type in {FileType.JSON, FileType.YAML, FileType.TOML}:
                    raw = self.load_file()
                    if raw is not None:
                        preview_str = (
                            json.dumps(raw, indent=2)
                            if isinstance(raw, dict)
                            else str(raw)
                        )
                        preview_lines = preview_str.splitlines()[:10]
                        content_tree = tree.add("[dim]Parsed preview:[/]")
                        for line in preview_lines:
                            content_tree.add(f"[dim]{line}[/]")
            except Exception as e:
                tree.add(f"[{OneColors.DARK_RED_b}]❌ Error reading file:[/] {e}")

        if not parent:
            self.console.print(tree)

    def __str__(self) -> str:
        return f"LoadFileAction(file_path={self.file_path}, file_type={self.file_type})"



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""selection_action.py"""
from typing import Any

from prompt_toolkit import PromptSession
from rich.tree import Tree

from falyx.action.action_types import SelectionReturnType
from falyx.action.base_action import BaseAction
from falyx.context import ExecutionContext
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import HookType
from falyx.logger import logger
from falyx.selection import (
    SelectionOption,
    SelectionOptionMap,
    prompt_for_index,
    prompt_for_selection,
    render_selection_dict_table,
    render_selection_indexed_table,
)
from falyx.signals import CancelSignal
from falyx.themes import OneColors


class SelectionAction(BaseAction):
    """
    A selection action that prompts the user to select an option from a list or
    dictionary. The selected option is then returned as the result of the action.

    If return_key is True, the key of the selected option is returned instead of
    the value.
    """

    def __init__(
        self,
        name: str,
        selections: (
            list[str]
            | set[str]
            | tuple[str, ...]
            | dict[str, SelectionOption]
            | dict[str, Any]
        ),
        *,
        title: str = "Select an option",
        columns: int = 5,
        prompt_message: str = "Select > ",
        default_selection: str = "",
        number_selections: int | str = 1,
        separator: str = ",",
        allow_duplicates: bool = False,
        inject_last_result: bool = False,
        inject_into: str = "last_result",
        return_type: SelectionReturnType | str = "value",
        prompt_session: PromptSession | None = None,
        never_prompt: bool = False,
        show_table: bool = True,
    ):
        super().__init__(
            name,
            inject_last_result=inject_last_result,
            inject_into=inject_into,
            never_prompt=never_prompt,
        )
        # Setter normalizes to correct type, mypy can't infer that
        self.selections: list[str] | SelectionOptionMap = selections  # type: ignore[assignment]
        self.return_type: SelectionReturnType = self._coerce_return_type(return_type)
        self.title = title
        self.columns = columns
        self.prompt_session = prompt_session or PromptSession()
        self.default_selection = default_selection
        self.number_selections = number_selections
        self.separator = separator
        self.allow_duplicates = allow_duplicates
        self.prompt_message = prompt_message
        self.show_table = show_table

    @property
    def number_selections(self) -> int | str:
        return self._number_selections

    @number_selections.setter
    def number_selections(self, value: int | str):
        if isinstance(value, int) and value > 0:
            self._number_selections: int | str = value
        elif isinstance(value, str):
            if value not in ("*"):
                raise ValueError("number_selections string must be '*'")
            self._number_selections = value
        else:
            raise ValueError("number_selections must be a positive integer or '*'")

    def _coerce_return_type(
        self, return_type: SelectionReturnType | str
    ) -> SelectionReturnType:
        if isinstance(return_type, SelectionReturnType):
            return return_type
        return SelectionReturnType(return_type)

    @property
    def selections(self) -> list[str] | SelectionOptionMap:
        return self._selections

    @selections.setter
    def selections(
        self, value: list[str] | set[str] | tuple[str, ...] | dict[str, SelectionOption]
    ):
        if isinstance(value, (list, tuple, set)):
            self._selections: list[str] | SelectionOptionMap = list(value)
        elif isinstance(value, dict):
            som = SelectionOptionMap()
            if all(isinstance(key, str) for key in value) and all(
                not isinstance(value[key], SelectionOption) for key in value
            ):
                som.update(
                    {
                        str(index): SelectionOption(key, option)
                        for index, (key, option) in enumerate(value.items())
                    }
                )
            elif all(isinstance(key, str) for key in value) and all(
                isinstance(value[key], SelectionOption) for key in value
            ):
                som.update(value)
            else:
                raise ValueError("Invalid dictionary format. Keys must be strings")
            self._selections = som
        else:
            raise TypeError(
                "'selections' must be a list[str] or dict[str, SelectionOption], "
                f"got {type(value).__name__}"
            )

    def _find_cancel_key(self) -> str:
        """Find the cancel key in the selections."""
        if isinstance(self.selections, dict):
            for index in range(len(self.selections) + 1):
                if str(index) not in self.selections:
                    return str(index)
        return str(len(self.selections))

    @property
    def cancel_key(self) -> str:
        return self._cancel_key

    @cancel_key.setter
    def cancel_key(self, value: str) -> None:
        """Set the cancel key for the selection."""
        if not isinstance(value, str):
            raise TypeError("Cancel key must be a string.")
        if isinstance(self.selections, dict) and value in self.selections:
            raise ValueError(
                "Cancel key cannot be one of the selection keys. "
                f"Current selections: {self.selections}"
            )
        if isinstance(self.selections, list):
            if not value.isdigit() or int(value) > len(self.selections):
                raise ValueError(
                    "cancel_key must be a digit and not greater than the number of selections."
                )
        self._cancel_key = value

    def cancel_formatter(self, index: int, selection: str) -> str:
        """Format the cancel option for display."""
        if self.cancel_key == str(index):
            return f"[{index}] [{OneColors.DARK_RED}]Cancel[/]"
        return f"[{index}] {selection}"

    def get_infer_target(self) -> tuple[None, None]:
        return None, None

    def _get_result_from_keys(self, keys: str | list[str]) -> Any:
        if not isinstance(self.selections, dict):
            raise TypeError("Selections must be a dictionary to get result by keys.")
        if self.return_type == SelectionReturnType.KEY:
            result: Any = keys
        elif self.return_type == SelectionReturnType.VALUE:
            if isinstance(keys, list):
                result = [self.selections[key].value for key in keys]
            elif isinstance(keys, str):
                result = self.selections[keys].value
        elif self.return_type == SelectionReturnType.ITEMS:
            if isinstance(keys, list):
                result = {key: self.selections[key] for key in keys}
            elif isinstance(keys, str):
                result = {keys: self.selections[keys]}
        elif self.return_type == SelectionReturnType.DESCRIPTION:
            if isinstance(keys, list):
                result = [self.selections[key].description for key in keys]
            elif isinstance(keys, str):
                result = self.selections[keys].description
        elif self.return_type == SelectionReturnType.DESCRIPTION_VALUE:
            if isinstance(keys, list):
                result = {
                    self.selections[key].description: self.selections[key].value
                    for key in keys
                }
            elif isinstance(keys, str):
                result = {self.selections[keys].description: self.selections[keys].value}
        else:
            raise ValueError(f"Unsupported return type: {self.return_type}")
        return result

    async def _run(self, *args, **kwargs) -> Any:
        kwargs = self._maybe_inject_last_result(kwargs)
        context = ExecutionContext(
            name=self.name,
            args=args,
            kwargs=kwargs,
            action=self,
        )

        effective_default = str(self.default_selection)
        maybe_result = str(self.last_result)
        if isinstance(self.selections, dict):
            if maybe_result in self.selections:
                effective_default = maybe_result
            elif self.inject_last_result:
                logger.warning(
                    "[%s] Injected last result '%s' not found in selections",
                    self.name,
                    maybe_result,
                )
        elif isinstance(self.selections, list):
            if maybe_result.isdigit() and int(maybe_result) in range(
                len(self.selections)
            ):
                effective_default = maybe_result
            elif self.inject_last_result:
                logger.warning(
                    "[%s] Injected last result '%s' not found in selections",
                    self.name,
                    maybe_result,
                )

        if self.never_prompt and not effective_default:
            raise ValueError(
                f"[{self.name}] 'never_prompt' is True but no valid default_selection "
                "or usable last_result was available."
            )

        context.start_timer()
        try:
            self.cancel_key = self._find_cancel_key()
            await self.hooks.trigger(HookType.BEFORE, context)
            if isinstance(self.selections, list):
                table = render_selection_indexed_table(
                    title=self.title,
                    selections=self.selections + ["Cancel"],
                    columns=self.columns,
                    formatter=self.cancel_formatter,
                )
                if not self.never_prompt:
                    indices: int | list[int] = await prompt_for_index(
                        len(self.selections),
                        table,
                        default_selection=effective_default,
                        prompt_session=self.prompt_session,
                        prompt_message=self.prompt_message,
                        show_table=self.show_table,
                        number_selections=self.number_selections,
                        separator=self.separator,
                        allow_duplicates=self.allow_duplicates,
                        cancel_key=self.cancel_key,
                    )
                else:
                    if effective_default:
                        indices = int(effective_default)
                    else:
                        raise ValueError(
                            f"[{self.name}] 'never_prompt' is True but no valid "
                            "default_selection was provided."
                        )

                if indices == int(self.cancel_key):
                    raise CancelSignal("User cancelled the selection.")
                if isinstance(indices, list):
                    result: str | list[str] = [
                        self.selections[index] for index in indices
                    ]
                elif isinstance(indices, int):
                    result = self.selections[indices]
                else:
                    assert False, "unreachable"
            elif isinstance(self.selections, dict):
                cancel_option = {
                    self.cancel_key: SelectionOption(
                        description="Cancel", value=CancelSignal, style=OneColors.DARK_RED
                    )
                }
                table = render_selection_dict_table(
                    title=self.title,
                    selections=self.selections | cancel_option,
                    columns=self.columns,
                )
                if not self.never_prompt:
                    keys = await prompt_for_selection(
                        (self.selections | cancel_option).keys(),
                        table,
                        default_selection=effective_default,
                        prompt_session=self.prompt_session,
                        prompt_message=self.prompt_message,
                        show_table=self.show_table,
                        number_selections=self.number_selections,
                        separator=self.separator,
                        allow_duplicates=self.allow_duplicates,
                        cancel_key=self.cancel_key,
                    )
                else:
                    keys = effective_default
                if keys == self.cancel_key:
                    raise CancelSignal("User cancelled the selection.")

                result = self._get_result_from_keys(keys)
            else:
                raise TypeError(
                    "'selections' must be a list[str] or dict[str, Any], "
                    f"got {type(self.selections).__name__}"
                )
            context.result = result
            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return result
        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def preview(self, parent: Tree | None = None):
        label = f"[{OneColors.LIGHT_RED}]🧭 SelectionAction[/] '{self.name}'"
        tree = parent.add(label) if parent else Tree(label)

        if isinstance(self.selections, list):
            sub = tree.add(f"[dim]Type:[/] List[str] ({len(self.selections)} items)")
            for i, item in enumerate(self.selections[:10]):  # limit to 10
                sub.add(f"[dim]{i}[/]: {item}")
            if len(self.selections) > 10:
                sub.add(f"[dim]... ({len(self.selections) - 10} more)[/]")
        elif isinstance(self.selections, dict):
            sub = tree.add(
                f"[dim]Type:[/] Dict[str, (str, Any)] ({len(self.selections)} items)"
            )
            for i, (key, option) in enumerate(list(self.selections.items())[:10]):
                sub.add(f"[dim]{key}[/]: {option.description}")
            if len(self.selections) > 10:
                sub.add(f"[dim]... ({len(self.selections) - 10} more)[/]")
        else:
            tree.add(f"[{OneColors.DARK_RED_b}]Invalid selections type[/]")
            return

        tree.add(f"[dim]Default:[/] '{self.default_selection or self.last_result}'")
        tree.add(f"[dim]Return:[/] {self.return_type.name.capitalize()}")
        tree.add(f"[dim]Prompt:[/] {'Disabled' if self.never_prompt else 'Enabled'}")

        if not parent:
            self.console.print(tree)

    def __str__(self) -> str:
        selection_type = (
            "List"
            if isinstance(self.selections, list)
            else "Dict" if isinstance(self.selections, dict) else "Unknown"
        )
        return (
            f"SelectionAction(name={self.name!r}, type={selection_type}, "
            f"default_selection={self.default_selection!r}, "
            f"return_type={self.return_type!r}, "
            f"prompt={'off' if self.never_prompt else 'on'})"
        )



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""fallback_action.py"""
from functools import cached_property
from typing import Any

from rich.tree import Tree

from falyx.action.action import Action
from falyx.themes import OneColors


class FallbackAction(Action):
    """
    FallbackAction provides a default value if the previous action failed or
    returned None.

    It injects the last result and checks:
    - If last_result is not None, it passes it through unchanged.
    - If last_result is None (e.g., due to failure), it replaces it with a fallback value.

    Used in ChainedAction pipelines to gracefully recover from errors or missing data.
    When activated, it consumes the preceding error and allows the chain to continue
    normally.

    Args:
        fallback (Any): The fallback value to use if last_result is None.
    """

    def __init__(self, fallback: Any):
        self._fallback = fallback

        async def _fallback_logic(last_result):
            return last_result if last_result is not None else fallback

        super().__init__(name="Fallback", action=_fallback_logic, inject_last_result=True)

    @cached_property
    def fallback(self) -> Any:
        """Return the fallback value."""
        return self._fallback

    async def preview(self, parent: Tree | None = None):
        label = [f"[{OneColors.LIGHT_RED}]🛟 Fallback[/] '{self.name}'"]
        label.append(f" [dim](uses fallback = {repr(self.fallback)})[/dim]")
        if parent:
            parent.add("".join(label))
        else:
            self.console.print(Tree("".join(label)))

    def __str__(self) -> str:
        return f"FallbackAction(fallback={self.fallback!r})"



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""shell_action.py
Execute shell commands with input substitution."""

from __future__ import annotations

import shlex
import subprocess
import sys
from typing import Any, Callable

from rich.tree import Tree

from falyx.action.io_action import BaseIOAction
from falyx.exceptions import FalyxError
from falyx.themes import OneColors


class ShellAction(BaseIOAction):
    """
    ShellAction wraps a shell command template for CLI pipelines.

    This Action takes parsed input (from stdin, literal, or last_result),
    substitutes it into the provided shell command template, and executes
    the command asynchronously using subprocess.

    Designed for quick integration with shell tools like `grep`, `ping`, `jq`, etc.

    ⚠️ Security Warning:
    By default, ShellAction uses `shell=True`, which can be dangerous with
    unsanitized input. To mitigate this, set `safe_mode=True` to use `shell=False`
    with `shlex.split()`.

    Features:
    - Automatically handles input parsing (str/bytes)
    - `safe_mode=True` disables shell interpretation and runs with `shell=False`
    - Captures stdout and stderr from shell execution
    - Raises on non-zero exit codes with stderr as the error
    - Result is returned as trimmed stdout string

    Args:
        name (str): Name of the action.
        command_template (str): Shell command to execute. Must include `{}` to include
                                input. If no placeholder is present, the input is not
                                included.
        safe_mode (bool): If True, runs with `shell=False` using shlex parsing
                          (default: False).
    """

    def __init__(
        self, name: str, command_template: str, safe_mode: bool = False, **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.command_template = command_template
        self.safe_mode = safe_mode

    def from_input(self, raw: str | bytes) -> str:
        if not isinstance(raw, (str, bytes)):
            raise TypeError(
                f"{self.name} expected str or bytes input, got {type(raw).__name__}"
            )
        return raw.strip() if isinstance(raw, str) else raw.decode("utf-8").strip()

    def get_infer_target(self) -> tuple[Callable[..., Any] | None, dict[str, Any] | None]:
        if sys.stdin.isatty():
            return self._run, {"parsed_input": {"help": self.command_template}}
        return None, None

    async def _run(self, parsed_input: str) -> str:
        # Replace placeholder in template, or use raw input as full command
        command = self.command_template.format(parsed_input)
        if self.safe_mode:
            try:
                args = shlex.split(command)
            except ValueError as error:
                raise FalyxError(f"Invalid command template: {error}")
            result = subprocess.run(args, capture_output=True, text=True, check=True)
        else:
            result = subprocess.run(
                command, shell=True, text=True, capture_output=True, check=True
            )
        if result.returncode != 0:
            raise RuntimeError(result.stderr.strip())
        return result.stdout.strip()

    def to_output(self, result: str) -> str:
        return result

    async def preview(self, parent: Tree | None = None):
        label = [f"[{OneColors.GREEN_b}]⚙ ShellAction[/] '{self.name}'"]
        label.append(f"\n[dim]Template:[/] {self.command_template}")
        label.append(
            f"\n[dim]Safe mode:[/] {'Enabled' if self.safe_mode else 'Disabled'}"
        )
        if self.inject_last_result:
            label.append(f" [dim](injects '{self.inject_into}')[/dim]")
        tree = parent.add("".join(label)) if parent else Tree("".join(label))
        if not parent:
            self.console.print(tree)

    def __str__(self):
        return (
            f"ShellAction(name={self.name!r}, command_template={self.command_template!r},"
            f" safe_mode={self.safe_mode})"
        )



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""config.py
Configuration loader for Falyx CLI commands."""
from __future__ import annotations

import importlib
import sys
from pathlib import Path
from typing import Any, Callable

import toml
import yaml
from pydantic import BaseModel, Field, field_validator, model_validator

from falyx.action.action import Action
from falyx.action.base_action import BaseAction
from falyx.command import Command
from falyx.console import console
from falyx.falyx import Falyx
from falyx.logger import logger
from falyx.retry import RetryPolicy
from falyx.themes import OneColors


def wrap_if_needed(obj: Any, name=None) -> BaseAction | Command:
    if isinstance(obj, (BaseAction, Command)):
        return obj
    elif callable(obj):
        return Action(name=name or getattr(obj, "__name__", "unnamed"), action=obj)
    else:
        raise TypeError(
            f"Cannot wrap object of type '{type(obj).__name__}'. "
            "Expected a function or BaseAction."
        )


def import_action(dotted_path: str) -> Any:
    """Dynamically imports a callable from a dotted path like 'my.module.func'."""
    module_path, _, attr = dotted_path.rpartition(".")
    if not module_path:
        console.print(f"[{OneColors.DARK_RED}]❌ Invalid action path:[/] {dotted_path}")
        sys.exit(1)
    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError as error:
        logger.error("Failed to import module '%s': %s", module_path, error)
        console.print(
            f"[{OneColors.DARK_RED}]❌ Could not import '{dotted_path}': {error}[/]\n"
            f"[{OneColors.COMMENT_GREY}]Ensure the module is installed and discoverable "
            "via PYTHONPATH."
        )
        sys.exit(1)
    try:
        action = getattr(module, attr)
    except AttributeError as error:
        logger.error(
            "Module '%s' does not have attribute '%s': %s", module_path, attr, error
        )
        console.print(
            f"[{OneColors.DARK_RED}]❌ Module '{module_path}' has no attribute "
            f"'{attr}': {error}[/]"
        )
        sys.exit(1)
    return action


class RawCommand(BaseModel):
    """Raw command model for Falyx CLI configuration."""

    key: str
    description: str
    action: str

    args: tuple[Any, ...] = Field(default_factory=tuple)
    kwargs: dict[str, Any] = Field(default_factory=dict)
    aliases: list[str] = Field(default_factory=list)
    tags: list[str] = Field(default_factory=list)
    style: str = OneColors.WHITE

    confirm: bool = False
    confirm_message: str = "Are you sure?"
    preview_before_confirm: bool = True

    spinner: bool = False
    spinner_message: str = "Processing..."
    spinner_type: str = "dots"
    spinner_style: str = OneColors.CYAN
    spinner_kwargs: dict[str, Any] = Field(default_factory=dict)

    before_hooks: list[Callable] = Field(default_factory=list)
    success_hooks: list[Callable] = Field(default_factory=list)
    error_hooks: list[Callable] = Field(default_factory=list)
    after_hooks: list[Callable] = Field(default_factory=list)
    teardown_hooks: list[Callable] = Field(default_factory=list)

    logging_hooks: bool = False
    retry: bool = False
    retry_all: bool = False
    retry_policy: RetryPolicy = Field(default_factory=RetryPolicy)
    hidden: bool = False
    help_text: str = ""
    help_epilog: str = ""

    @field_validator("retry_policy")
    @classmethod
    def validate_retry_policy(cls, value: dict | RetryPolicy) -> RetryPolicy:
        if isinstance(value, RetryPolicy):
            return value
        if not isinstance(value, dict):
            raise ValueError("retry_policy must be a dictionary.")
        return RetryPolicy(**value)


def convert_commands(raw_commands: list[dict[str, Any]]) -> list[Command]:
    commands = []
    for entry in raw_commands:
        raw_command = RawCommand(**entry)
        commands.append(
            Command.model_validate(
                {
                    **raw_command.model_dump(exclude={"action"}),
                    "action": wrap_if_needed(
                        import_action(raw_command.action), name=raw_command.description
                    ),
                }
            )
        )

    return commands


def convert_submenus(
    raw_submenus: list[dict[str, Any]], *, parent_path: Path | None = None, depth: int = 0
) -> list[dict[str, Any]]:
    submenus: list[dict[str, Any]] = []
    for raw_submenu in raw_submenus:
        if raw_submenu.get("config"):
            config_path = Path(raw_submenu["config"])
            if parent_path:
                config_path = (parent_path.parent / config_path).resolve()
            submenu = loader(config_path, _depth=depth + 1)
        else:
            submenu_module_path = raw_submenu.get("submenu")
            if not isinstance(submenu_module_path, str):
                console.print(
                    f"[{OneColors.DARK_RED}]❌ Invalid submenu path:[/] {submenu_module_path}"
                )
                sys.exit(1)
            submenu = import_action(submenu_module_path)
        if not isinstance(submenu, Falyx):
            console.print(f"[{OneColors.DARK_RED}]❌ Invalid submenu:[/] {submenu}")
            sys.exit(1)

        key = raw_submenu.get("key")
        if not isinstance(key, str):
            console.print(f"[{OneColors.DARK_RED}]❌ Invalid submenu key:[/] {key}")
            sys.exit(1)

        description = raw_submenu.get("description")
        if not isinstance(description, str):
            console.print(
                f"[{OneColors.DARK_RED}]❌ Invalid submenu description:[/] {description}"
            )
            sys.exit(1)

        submenus.append(
            Submenu(
                key=key,
                description=description,
                submenu=submenu,
                style=raw_submenu.get("style", OneColors.CYAN),
            ).model_dump()
        )
    return submenus


class Submenu(BaseModel):
    """Submenu model for Falyx CLI configuration."""

    key: str
    description: str
    submenu: Any
    style: str = OneColors.CYAN


class FalyxConfig(BaseModel):
    """Falyx CLI configuration model."""

    title: str = "Falyx CLI"
    prompt: str | list[tuple[str, str]] | list[list[str]] = [
        (OneColors.BLUE_b, "FALYX > ")
    ]
    columns: int = 4
    welcome_message: str = ""
    exit_message: str = ""
    commands: list[Command] | list[dict] = []
    submenus: list[dict[str, Any]] = []

    @model_validator(mode="after")
    def validate_prompt_format(self) -> FalyxConfig:
        if isinstance(self.prompt, list):
            for pair in self.prompt:
                if not isinstance(pair, (list, tuple)) or len(pair) != 2:
                    raise ValueError(
                        "Prompt list must contain 2-element (style, text) pairs"
                    )
        return self

    def to_falyx(self) -> Falyx:
        flx = Falyx(
            title=self.title,
            prompt=self.prompt,  # type: ignore[arg-type]
            columns=self.columns,
            welcome_message=self.welcome_message,
            exit_message=self.exit_message,
        )
        flx.add_commands(self.commands)
        for submenu in self.submenus:
            flx.add_submenu(**submenu)
        return flx


def loader(file_path: Path | str, _depth: int = 0) -> Falyx:
    """
    Load Falyx CLI configuration from a YAML or TOML file.

    The file should contain a dictionary with a list of commands.

    Each command should be defined as a dictionary with at least:
    - key: a unique single-character key
    - description: short description
    - action: dotted import path to the action function/class

    Args:
        file_path (str): Path to the config file (YAML or TOML).

    Returns:
        Falyx: An instance of the Falyx CLI with loaded commands.

    Raises:
        ValueError: If the file format is unsupported or file cannot be parsed.
    """
    if _depth > 5:
        raise ValueError("Maximum submenu depth exceeded (5 levels deep)")

    if isinstance(file_path, (str, Path)):
        path = Path(file_path)
    else:
        raise TypeError("file_path must be a string or Path object.")

    if not path.is_file():
        raise FileNotFoundError(f"No such config file: {file_path}")

    suffix = path.suffix
    with path.open("r", encoding="UTF-8") as config_file:
        if suffix in (".yaml", ".yml"):
            raw_config = yaml.safe_load(config_file)
        elif suffix == ".toml":
            raw_config = toml.load(config_file)
        else:
            raise ValueError(f"Unsupported config format: {suffix}")

    if not isinstance(raw_config, dict):
        raise ValueError(
            "Configuration file must contain a dictionary with a list of commands.\n"
            "Example:\n"
            "title: 'My CLI'\n"
            "commands:\n"
            "  - key: 'a'\n"
            "    description: 'Example command'\n"
            "    action: 'my_module.my_function'"
        )

    commands = convert_commands(raw_config["commands"])
    submenus = convert_submenus(raw_config.get("submenus", []))
    return FalyxConfig(
        title=raw_config.get("title", f"[{OneColors.BLUE_b}]Falyx CLI"),
        prompt=raw_config.get("prompt", [(OneColors.BLUE_b, "FALYX > ")]),
        columns=raw_config.get("columns", 4),
        welcome_message=raw_config.get("welcome_message", ""),
        exit_message=raw_config.get("exit_message", ""),
        commands=commands,
        submenus=submenus,
    ).to_falyx()



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""save_file_action.py"""
import csv
import json
import xml.etree.ElementTree as ET
from datetime import datetime
from pathlib import Path
from typing import Any, Literal

import toml
import yaml
from rich.tree import Tree

from falyx.action.action_types import FileType
from falyx.action.base_action import BaseAction
from falyx.context import ExecutionContext
from falyx.execution_registry import ExecutionRegistry as er
from falyx.hook_manager import HookType
from falyx.logger import logger
from falyx.themes import OneColors


class SaveFileAction(BaseAction):
    """
    SaveFileAction saves data to a file in the specified format (e.g., TEXT, JSON, YAML).
    Supports overwrite control and integrates with chaining workflows via inject_last_result.

    Supported types: TEXT, JSON, YAML, TOML, CSV, TSV, XML

    If the file exists and overwrite is False, the action will raise a FileExistsError.
    """

    def __init__(
        self,
        name: str,
        file_path: str,
        file_type: FileType | str = FileType.TEXT,
        mode: Literal["w", "a"] = "w",
        data: Any = None,
        overwrite: bool = True,
        create_dirs: bool = True,
        inject_last_result: bool = False,
        inject_into: str = "data",
    ):
        """
        SaveFileAction allows saving data to a file.

        Args:
            name (str): Name of the action.
            file_path (str | Path): Path to the file where data will be saved.
            file_type (FileType | str): Format to write to (e.g. TEXT, JSON, YAML).
            mode (Literal["w", "a"]): File mode (default: "w").
            data (Any): Data to be saved (if not using inject_last_result).
            overwrite (bool): Whether to overwrite the file if it exists.
            create_dirs (bool): Whether to create parent directories if they do not exist.
            inject_last_result (bool): Whether to inject result from previous action.
            inject_into (str): Kwarg name to inject the last result as.
        """
        super().__init__(
            name=name, inject_last_result=inject_last_result, inject_into=inject_into
        )
        self._file_path = self._coerce_file_path(file_path)
        self._file_type = self._coerce_file_type(file_type)
        self.data = data
        self.overwrite = overwrite
        self.mode = mode
        self.create_dirs = create_dirs

    @property
    def file_path(self) -> Path | None:
        """Get the file path as a Path object."""
        return self._file_path

    @file_path.setter
    def file_path(self, value: str | Path):
        """Set the file path, converting to Path if necessary."""
        self._file_path = self._coerce_file_path(value)

    def _coerce_file_path(self, file_path: str | Path | None) -> Path | None:
        """Coerce the file path to a Path object."""
        if isinstance(file_path, Path):
            return file_path
        elif isinstance(file_path, str):
            return Path(file_path)
        elif file_path is None:
            return None
        else:
            raise TypeError("file_path must be a string or Path object")

    @property
    def file_type(self) -> FileType:
        """Get the file type."""
        return self._file_type

    @file_type.setter
    def file_type(self, value: FileType | str):
        """Set the file type, converting to FileType if necessary."""
        self._file_type = self._coerce_file_type(value)

    def _coerce_file_type(self, file_type: FileType | str) -> FileType:
        """Coerce the file type to a FileType enum."""
        if isinstance(file_type, FileType):
            return file_type
        elif isinstance(file_type, str):
            return FileType(file_type)
        else:
            raise TypeError("file_type must be a FileType enum or string")

    def get_infer_target(self) -> tuple[None, None]:
        return None, None

    def _dict_to_xml(self, data: dict, root: ET.Element) -> None:
        """Convert a dictionary to XML format."""
        for key, value in data.items():
            if isinstance(value, dict):
                sub_element = ET.SubElement(root, key)
                self._dict_to_xml(value, sub_element)
            elif isinstance(value, list):
                for item in value:
                    item_element = ET.SubElement(root, key)
                    if isinstance(item, dict):
                        self._dict_to_xml(item, item_element)
                    else:
                        item_element.text = str(item)
            else:
                element = ET.SubElement(root, key)
                element.text = str(value)

    async def save_file(self, data: Any) -> None:
        """Save data to the specified file in the desired format."""
        if self.file_path is None:
            raise ValueError("file_path must be set before saving a file")
        elif self.file_path.exists() and not self.overwrite:
            raise FileExistsError(f"File already exists: {self.file_path}")

        if self.file_path.parent and not self.file_path.parent.exists():
            if self.create_dirs:
                self.file_path.parent.mkdir(parents=True, exist_ok=True)
            else:
                raise FileNotFoundError(
                    f"Directory does not exist: {self.file_path.parent}"
                )

        try:
            if self.file_type == FileType.TEXT:
                self.file_path.write_text(data, encoding="UTF-8")
            elif self.file_type == FileType.JSON:
                self.file_path.write_text(json.dumps(data, indent=4), encoding="UTF-8")
            elif self.file_type == FileType.TOML:
                self.file_path.write_text(toml.dumps(data), encoding="UTF-8")
            elif self.file_type == FileType.YAML:
                self.file_path.write_text(yaml.dump(data), encoding="UTF-8")
            elif self.file_type == FileType.CSV:
                if not isinstance(data, list) or not all(
                    isinstance(row, list) for row in data
                ):
                    raise ValueError(
                        f"{self.file_type.name} file type requires a list of lists"
                    )
                with open(
                    self.file_path, mode=self.mode, newline="", encoding="UTF-8"
                ) as csvfile:
                    writer = csv.writer(csvfile)
                    writer.writerows(data)
            elif self.file_type == FileType.TSV:
                if not isinstance(data, list) or not all(
                    isinstance(row, list) for row in data
                ):
                    raise ValueError(
                        f"{self.file_type.name} file type requires a list of lists"
                    )
                with open(
                    self.file_path, mode=self.mode, newline="", encoding="UTF-8"
                ) as tsvfile:
                    writer = csv.writer(tsvfile, delimiter="\t")
                    writer.writerows(data)
            elif self.file_type == FileType.XML:
                if not isinstance(data, dict):
                    raise ValueError("XML file type requires data to be a dictionary")
                root = ET.Element("root")
                self._dict_to_xml(data, root)
                tree = ET.ElementTree(root)
                tree.write(self.file_path, encoding="UTF-8", xml_declaration=True)
            else:
                raise ValueError(f"Unsupported file type: {self.file_type}")

        except Exception as error:
            logger.error("Failed to save %s: %s", self.file_path.name, error)
            raise

    async def _run(self, *args, **kwargs):
        combined_kwargs = self._maybe_inject_last_result(kwargs)
        data = self.data or combined_kwargs.get(self.inject_into)

        context = ExecutionContext(
            name=self.name, args=args, kwargs=combined_kwargs, action=self
        )
        context.start_timer()

        try:
            await self.hooks.trigger(HookType.BEFORE, context)

            await self.save_file(data)
            logger.debug("File saved successfully: %s", self.file_path)

            await self.hooks.trigger(HookType.ON_SUCCESS, context)
            return str(self.file_path)

        except Exception as error:
            context.exception = error
            await self.hooks.trigger(HookType.ON_ERROR, context)
            raise
        finally:
            context.stop_timer()
            await self.hooks.trigger(HookType.AFTER, context)
            await self.hooks.trigger(HookType.ON_TEARDOWN, context)
            er.record(context)

    async def preview(self, parent: Tree | None = None):
        label = f"[{OneColors.CYAN}]💾 SaveFileAction[/] '{self.name}'"
        tree = parent.add(label) if parent else Tree(label)

        tree.add(f"[dim]Path:[/] {self.file_path}")
        tree.add(f"[dim]Type:[/] {self.file_type.name}")
        tree.add(f"[dim]Overwrite:[/] {self.overwrite}")

        if self.file_path and self.file_path.exists():
            if self.overwrite:
                tree.add(f"[{OneColors.LIGHT_YELLOW}]⚠️ File will be overwritten[/]")
            else:
                tree.add(
                    f"[{OneColors.DARK_RED}]❌ File exists and overwrite is disabled[/]"
                )
            stat = self.file_path.stat()
            tree.add(f"[dim]Size:[/] {stat.st_size:,} bytes")
            tree.add(
                f"[dim]Modified:[/] {datetime.fromtimestamp(stat.st_mtime):%Y-%m-%d %H:%M:%S}"
            )
            tree.add(
                f"[dim]Created:[/] {datetime.fromtimestamp(stat.st_ctime):%Y-%m-%d %H:%M:%S}"
            )

        if not parent:
            self.console.print(tree)

    def __str__(self) -> str:
        return f"SaveFileAction(file_path={self.file_path}, file_type={self.file_type})"



from collections import Counter
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.document import Document
from typing import Iterable, Set, Optional
import shlex

from falyx.command import Command
from falyx.parser.command_argument_parser import CommandArgumentParser
from falyx.parser.argument import Argument
from falyx.parser.argument_action import ArgumentAction

class FalyxCompleter(Completer):
    """Completer for Falyx commands and their arguments."""
    def __init__(self, falyx: "Falyx"):
        self.falyx = falyx
        self._used_args: Set[str] = set()
        self._used_args_counter: Counter = Counter()

    def get_completions(self, document: Document, complete_event) -> Iterable[Completion]:
        text = document.text_before_cursor
        try:
            tokens = shlex.split(text)
            cursor_at_end_of_token = document.text_before_cursor.endswith((' ', '\t'))
        except ValueError:
            return

        if not tokens or (len(tokens) == 1 and not cursor_at_end_of_token):
            # Suggest command keys and aliases
            yield from self._suggest_commands(tokens[0] if tokens else "")
            return

        command = self._match_command(tokens[0])
        if not command:
            return

        if command.arg_parser is None:
            return

        self._set_used_args(tokens, command)

        next_arg = self._next_expected_argument(tokens, command.arg_parser)

        if next_arg:
            # Positional arguments or required flagged arguments
            yield from self._suggest_argument(next_arg, document)
        else:
            # Optional arguments
            for arg in command.arg_parser._keyword.values():
                if not self._arg_already_used(arg.dest):
                    yield from self._suggest_argument(arg, document)

    def _set_used_args(self, tokens: list[str], command: Command) -> None:
        """Extracts used argument flags from the provided tokens."""
        if not command.arg_parser:
            return
        self._used_args.clear()
        self._used_args_counter.clear()
        for token in tokens[1:]:
            if token.startswith('-'):
                if keyword_argument := command.arg_parser._keyword.get(token):
                    self._used_args_counter[keyword_argument.dest] += 1
                    if isinstance(keyword_argument.nargs, int) and self._used_args_counter[keyword_argument.dest] > keyword_argument.nargs:
                        continue
                    elif isinstance(keyword_argument.nargs, str) and keyword_argument.nargs in ("?"):
                        self._used_args.add(keyword_argument.dest)
                    else:
                        self._used_args.add(keyword_argument.dest)
            else:
                # Handle positional arguments
                if command.arg_parser._positional:
                    for arg in command.arg_parser._positional.values():
                        if arg.dest not in self._used_args:
                            self._used_args.add(arg.dest)
                            break
        print(f"Used args: {self._used_args}, Counter: {self._used_args_counter}")

    def _suggest_commands(self, prefix: str) -> Iterable[Completion]:
        prefix = prefix.upper()
        seen = set()
        for cmd in self.falyx.commands.values():
            for key in [cmd.key] + cmd.aliases:
                if key.upper().startswith(prefix) and key not in seen:
                    yield Completion(key, start_position=-len(prefix))
                    seen.add(key)

    def _match_command(self, token: str) -> Optional[Command]:
        token = token.lstrip("?").upper()
        return self.falyx._name_map.get(token)

    def _next_expected_argument(
        self, tokens: list[str], parser: CommandArgumentParser
    ) -> Optional[Argument]:
        """Determine the next expected argument based on the current tokens."""
        # Positional arguments first
        for arg in parser._positional.values():
            if arg.dest not in self._used_args:
                return arg

        # Then required keyword arguments
        for arg in parser._keyword_list:
            if arg.required and not self._arg_already_used(arg.dest):
                return arg

        return None

    def _arg_already_used(self, dest: str) -> bool:
        print(f"Checking if argument '{dest}' is already used: {dest in self._used_args} - Used args: {self._used_args}")
        return dest in self._used_args

    def _suggest_argument(self, arg: Argument, document: Document) -> Iterable[Completion]:
        if not arg.positional:
            for flag in arg.flags:
                yield Completion(flag, start_position=0)

        if arg.choices:
            for choice in arg.choices:
                yield Completion(
                    choice,
                    start_position=0,
                    display=f"{arg.dest}={choice}"
                )

        if arg.default is not None and arg.action == ArgumentAction.STORE:
            yield Completion(
                str(arg.default),
                start_position=0,
                display=f"{arg.dest} (default: {arg.default})"
            )



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""parser_types.py"""
from typing import Any


def true_none(value: Any) -> bool | None:
    if value is None:
        return None
    return True


def false_none(value: Any) -> bool | None:
    if value is None:
        return None
    return False



# Falyx CLI Framework — (c) 2025 rtj.dev LLC — MIT Licensed
"""selection.py"""
from dataclasses import dataclass
from typing import Any, Callable, KeysView, Sequence

from prompt_toolkit import PromptSession
from rich import box
from rich.markup import escape
from rich.table import Table

from falyx.console import console
from falyx.themes import OneColors
from falyx.utils import CaseInsensitiveDict, chunks
from falyx.validators import MultiIndexValidator, MultiKeyValidator


@dataclass
class SelectionOption:
    """Represents a single selection option with a description and a value."""

    description: str
    value: Any
    style: str = OneColors.WHITE

    def __post_init__(self):
        if not isinstance(self.description, str):
            raise TypeError("SelectionOption description must be a string.")

    def render(self, key: str) -> str:
        """Render the selection option for display."""
        key = escape(f"[{key}]")
        return f"[{OneColors.WHITE}]{key}[/] [{self.style}]{self.description}[/]"


class SelectionOptionMap(CaseInsensitiveDict):
    """
    Manages selection options including validation and reserved key protection.
    """

    RESERVED_KEYS: set[str] = set()

    def __init__(
        self,
        options: dict[str, SelectionOption] | None = None,
        allow_reserved: bool = False,
    ):
        super().__init__()
        self.allow_reserved = allow_reserved
        if options:
            self.update(options)

    def _add_reserved(self, key: str, option: SelectionOption) -> None:
        """Add a reserved key, bypassing validation."""
        norm_key = key.upper()
        super().__setitem__(norm_key, option)

    def __setitem__(self, key: str, option: SelectionOption) -> None:
        if not isinstance(option, SelectionOption):
            raise TypeError(f"Value for key '{key}' must be a SelectionOption.")
        norm_key = key.upper()
        if norm_key in self.RESERVED_KEYS and not self.allow_reserved:
            raise ValueError(
                f"Key '{key}' is reserved and cannot be used in SelectionOptionMap."
            )
        super().__setitem__(norm_key, option)

    def __delitem__(self, key: str) -> None:
        if key.upper() in self.RESERVED_KEYS and not self.allow_reserved:
            raise ValueError(f"Cannot delete reserved option '{key}'.")
        super().__delitem__(key)

    def update(self, other=None, **kwargs):
        """Update the selection options with another dictionary."""
        if other:
            for key, option in other.items():
                if not isinstance(option, SelectionOption):
                    raise TypeError(f"Value for key '{key}' must be a SelectionOption.")
                self[key] = option
        for key, option in kwargs.items():
            if not isinstance(option, SelectionOption):
                raise TypeError(f"Value for key '{key}' must be a SelectionOption.")
            self[key] = option

    def items(self, include_reserved: bool = True):
        for k, v in super().items():
            if not include_reserved and k in self.RESERVED_KEYS:
                continue
            yield k, v


def render_table_base(
    title: str,
    *,
    caption: str = "",
    columns: int = 4,
    box_style: box.Box = box.SIMPLE,
    show_lines: bool = False,
    show_header: bool = False,
    show_footer: bool = False,
    style: str = "",
    header_style: str = "",
    footer_style: str = "",
    title_style: str = "",
    caption_style: str = "",
    highlight: bool = True,
    column_names: Sequence[str] | None = None,
) -> Table:
    table = Table(
        title=title,
        caption=caption,
        box=box_style,
        show_lines=show_lines,
        show_header=show_header,
        show_footer=show_footer,
        style=style,
        header_style=header_style,
        footer_style=footer_style,
        title_style=title_style,
        caption_style=caption_style,
        highlight=highlight,
    )
    if column_names:
        for column_name in column_names:
            table.add_column(column_name)
    else:
        for _ in range(columns):
            table.add_column()
    return table


def render_selection_grid(
    title: str,
    selections: Sequence[str],
    *,
    columns: int = 4,
    caption: str = "",
    box_style: box.Box = box.SIMPLE,
    show_lines: bool = False,
    show_header: bool = False,
    show_footer: bool = False,
    style: str = "",
    header_style: str = "",
    footer_style: str = "",
    title_style: str = "",
    caption_style: str = "",
    highlight: bool = False,
) -> Table:
    """Create a selection table with the given parameters."""
    table = render_table_base(
        title=title,
        caption=caption,
        columns=columns,
        box_style=box_style,
        show_lines=show_lines,
        show_header=show_header,
        show_footer=show_footer,
        style=style,
        header_style=header_style,
        footer_style=footer_style,
        title_style=title_style,
        caption_style=caption_style,
        highlight=highlight,
    )

    for chunk in chunks(selections, columns):
        table.add_row(*chunk)

    return table


def render_selection_indexed_table(
    title: str,
    selections: Sequence[str],
    *,
    columns: int = 4,
    caption: str = "",
    box_style: box.Box = box.SIMPLE,
    show_lines: bool = False,
    show_header: bool = False,
    show_footer: bool = False,
    style: str = "",
    header_style: str = "",
    footer_style: str = "",
    title_style: str = "",
    caption_style: str = "",
    highlight: bool = False,
    formatter: Callable[[int, str], str] | None = None,
) -> Table:
    """Create a selection table with the given parameters."""
    table = render_table_base(
        title=title,
        caption=caption,
        columns=columns,
        box_style=box_style,
        show_lines=show_lines,
        show_header=show_header,
        show_footer=show_footer,
        style=style,
        header_style=header_style,
        footer_style=footer_style,
        title_style=title_style,
        caption_style=caption_style,
        highlight=highlight,
    )

    for indexes, chunk in zip(
        chunks(range(len(selections)), columns), chunks(selections, columns)
    ):
        row = [
            formatter(index, selection) if formatter else f"[{index}] {selection}"
            for index, selection in zip(indexes, chunk)
        ]
        table.add_row(*row)

    return table


def render_selection_dict_table(
    title: str,
    selections: dict[str, SelectionOption],
    *,
    columns: int = 2,
    caption: str = "",
    box_style: box.Box = box.SIMPLE,
    show_lines: bool = False,
    show_header: bool = False,
    show_footer: bool = False,
    style: str = "",
    header_style: str = "",
    footer_style: str = "",
    title_style: str = "",
    caption_style: str = "",
    highlight: bool = False,
) -> Table:
    """Create a selection table with the given parameters."""
    table = render_table_base(
        title=title,
        caption=caption,
        columns=columns,
        box_style=box_style,
        show_lines=show_lines,
        show_header=show_header,
        show_footer=show_footer,
        style=style,
        header_style=header_style,
        footer_style=footer_style,
        title_style=title_style,
        caption_style=caption_style,
        highlight=highlight,
    )

    for chunk in chunks(selections.items(), columns):
        row = []
        for key, option in chunk:
            row.append(
                f"[{OneColors.WHITE}][{key.upper()}] "
                f"[{option.style}]{option.description}[/]"
            )
        table.add_row(*row)

    return table


async def prompt_for_index(
    max_index: int,
    table: Table,
    *,
    min_index: int = 0,
    default_selection: str = "",
    prompt_session: PromptSession | None = None,
    prompt_message: str = "Select an option > ",
    show_table: bool = True,
    number_selections: int | str = 1,
    separator: str = ",",
    allow_duplicates: bool = False,
    cancel_key: str = "",
) -> int | list[int]:
    prompt_session = prompt_session or PromptSession()

    if show_table:
        console.print(table, justify="center")

    selection = await prompt_session.prompt_async(
        message=prompt_message,
        validator=MultiIndexValidator(
            min_index,
            max_index,
            number_selections,
            separator,
            allow_duplicates,
            cancel_key,
        ),
        default=default_selection,
    )

    if selection.strip() == cancel_key:
        return int(cancel_key)
    if isinstance(number_selections, int) and number_selections == 1:
        return int(selection.strip())
    return [int(index.strip()) for index in selection.strip().split(separator)]


async def prompt_for_selection(
    keys: Sequence[str] | KeysView[str],
    table: Table,
    *,
    default_selection: str = "",
    prompt_session: PromptSession | None = None,
    prompt_message: str = "Select an option > ",
    show_table: bool = True,
    number_selections: int | str = 1,
    separator: str = ",",
    allow_duplicates: bool = False,
    cancel_key: str = "",
) -> str | list[str]:
    """Prompt the user to select a key from a set of options. Return the selected key."""
    prompt_session = prompt_session or PromptSession()

    if show_table:
        console.print(table, justify="center")

    selected = await prompt_session.prompt_async(
        message=prompt_message,
        validator=MultiKeyValidator(
            keys, number_selections, separator, allow_duplicates, cancel_key
        ),
        default=default_selection,
    )

    if selected.strip() == cancel_key:
        return cancel_key
    if isinstance(number_selections, int) and number_selections == 1:
        return selected.strip()
    return [key.strip() for key in selected.strip().split(separator)]


async def select_value_from_list(
    title: str,
    selections: Sequence[str],
    *,
    prompt_session: PromptSession | None = None,
    prompt_message: str = "Select an option > ",
    default_selection: str = "",
    number_selections: int | str = 1,
    separator: str = ",",
    allow_duplicates: bool = False,
    cancel_key: str = "",
    columns: int = 4,
    caption: str = "",
    box_style: box.Box = box.SIMPLE,
    show_lines: bool = False,
    show_header: bool = False,
    show_footer: bool = False,
    style: str = "",
    header_style: str = "",
    footer_style: str = "",
    title_style: str = "",
    caption_style: str = "",
    highlight: bool = False,
) -> str | list[str]:
    """Prompt for a selection. Return the selected item."""
    table = render_selection_indexed_table(
        title=title,
        selections=selections,
        columns=columns,
        caption=caption,
        box_style=box_style,
        show_lines=show_lines,
        show_header=show_header,
        show_footer=show_footer,
        style=style,
        header_style=header_style,
        footer_style=footer_style,
        title_style=title_style,
        caption_style=caption_style,
        highlight=highlight,
    )
    prompt_session = prompt_session or PromptSession()

    selection_index = await prompt_for_index(
        len(selections) - 1,
        table,
        default_selection=default_selection,
        prompt_session=prompt_session,
        prompt_message=prompt_message,
        number_selections=number_selections,
        separator=separator,
        allow_duplicates=allow_duplicates,
        cancel_key=cancel_key,
    )

    if isinstance(selection_index, list):
        return [selections[i] for i in selection_index]
    return selections[selection_index]


async def select_key_from_dict(
    selections: dict[str, SelectionOption],
    table: Table,
    *,
    prompt_session: PromptSession | None = None,
    prompt_message: str = "Select an option > ",
    default_selection: str = "",
    number_selections: int | str = 1,
    separator: str = ",",
    allow_duplicates: bool = False,
    cancel_key: str = "",
) -> str | list[str]:
    """Prompt for a key from a dict, returns the key."""
    prompt_session = prompt_session or PromptSession()

    console.print(table, justify="center")

    return await prompt_for_selection(
        selections.keys(),
        table,
        default_selection=default_selection,
        prompt_session=prompt_session,
        prompt_message=prompt_message,
        number_selections=number_selections,
        separator=separator,
        allow_duplicates=allow_duplicates,
        cancel_key=cancel_key,
    )


async def select_value_from_dict(
    selections: dict[str, SelectionOption],
    table: Table,
    *,
    prompt_session: PromptSession | None = None,
    prompt_message: str = "Select an option > ",
    default_selection: str = "",
    number_selections: int | str = 1,
    separator: str = ",",
    allow_duplicates: bool = False,
    cancel_key: str = "",
) -> Any | list[Any]:
    """Prompt for a key from a dict, but return the value."""
    prompt_session = prompt_session or PromptSession()

    console.print(table, justify="center")

    selection_key = await prompt_for_selection(
        selections.keys(),
        table,
        default_selection=default_selection,
        prompt_session=prompt_session,
        prompt_message=prompt_message,
        number_selections=number_selections,
        separator=separator,
        allow_duplicates=allow_duplicates,
        cancel_key=cancel_key,
    )

    if isinstance(selection_key, list):
        return [selections[key].value for key in selection_key]
    return selections[selection_key].value


async def get_selection_from_dict_menu(
    title: str,
    selections: dict[str, SelectionOption],
    *,
    prompt_session: PromptSession | None = None,
    prompt_message: str = "Select an option > ",
    default_selection: str = "",
    number_selections: int | str = 1,
    separator: str = ",",
    allow_duplicates: bool = False,
    cancel_key: str = "",
) -> Any | list[Any]:
    """Prompt for a key from a dict, but return the value."""
    table = render_selection_dict_table(
        title,
        selections,
    )

    return await select_value_from_dict(
        selections=selections,
        table=table,
        prompt_session=prompt_session,
        prompt_message=prompt_message,
        default_selection=default_selection,
        number_selections=number_selections,
        separator=separator,
        allow_duplicates=allow_duplicates,
        cancel_key=cancel_key,
    )
